"""
Used to define global helpers and variables for MSFT events.
"""

from pypanther.helpers.base import deep_get, deep_walk


def msft_graph_alert_context(event):
    return {
        "category": event.get("category", ""),
        "description": event.get("description", ""),
        "userStates": event.get("userStates", []),
        "fileStates": event.get("fileStates", []),
        "hostStates": event.get("hostStates", []),
    }


def m365_alert_context(event):
    return {
        "operation": event.get("Operation", ""),
        "organization_id": event.get("OrganizationId", ""),
        "client_ip": event.get("ClientIp", ""),
        "extended_properties": event.get("ExtendedProperties", []),
        "modified_properties": event.get("ModifiedProperties", []),
        "application": event.get("Application", ""),
        "actor": event.get("Actor", []),
    }


def azure_rule_context(event: dict):
    return {
        "operationName": event.get("operationName", default="<MISSING_OPERATION_NAME>"),
        "category": deep_get(event, "properties", "category", default="<MISSING_CATEGORY>"),
        "actor_id": deep_get(
            event,
            "properties",
            "initiatedBy",
            "user",
            "id",
            default="<MISSING_ACTOR_ID",
        ),
        "actor_upn": deep_get(
            event,
            "properties",
            "initiatedBy",
            "user",
            "userPrincipalName",
            default="<MISSING UPN>",
        ),
        "source_ip_address": deep_get(
            event,
            "properties",
            "initiatedBy",
            "user",
            "ipAddress",
            default="<MISSING_SOURCE_IP>",
        ),
        "target_id": deep_walk(
            event,
            "properties",
            "targetResources",
            "id",
            default="<MISSING_ACCOUNT_ID>",
        ),
        "target_name": deep_walk(
            event,
            "properties",
            "targetResources",
            "displayName",
            default="<MISSING UPN>",
        ),
    }


def get_target_name(event, target_type="User"):
    target_resources = deep_walk(event, "properties", "targetResources", default="")
    for resource in target_resources:
        if resource.get("type") == target_type:
            return resource.get("displayName", "NO DISPLAY NAME FOUND")

    return "NO DISPLAY NAME FOUND"


def azure_success(event):
    result = event.deep_get("properties", "result", default="")
    return result == "success"


def is_external_address(address, primary_domain, onmicrosoft_domain):
    """
    Check if an email address is external to the organization.

    Args:
    ----
        address (str): The email address or SMTP address to check
        primary_domain (str): The organization's primary domain (e.g. contoso.com)
        onmicrosoft_domain (str): The tenant domain (e.g. contoso.onmicrosoft.com)

    Returns:
    -------
        bool: True if the address is external, False if internal

    """
    if not address or (not primary_domain and not onmicrosoft_domain):
        return True

    # Clean up and normalize the address
    address = address.lower()
    if address.startswith("smtp:"):
        address = address[5:]

    # Check each address (might be multiple addresses separated by semicolon)
    for addr in address.split(";"):
        try:
            domain = addr.strip().split("@")[1].lower()
            # Skip if internal (matches onmicrosoft domain or primary domain/subdomain)
            if (onmicrosoft_domain and domain == onmicrosoft_domain.lower()) or (
                primary_domain and (domain == primary_domain or domain.endswith("." + primary_domain))
            ):
                continue
            return True
        except (IndexError, AttributeError):
            return True

    return False



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed


@panther_managed
class RetrieveSSOaccesstoken(Rule):
    id = "Retrieve.SSO.access.token-prototype"
    display_name = "SIGNAL - Retrieve SSO access token"
    create_alert = False
    log_types = [LogType.AWS_CLOUDTRAIL]
    default_severity = Severity.INFO

    def rule(self, event):
        return event.get("eventSource") == "sso.amazonaws.com" and event.get("eventName") == "CreateToken"

    tests = [
        RuleTest(
            name="Retrieve SSO access token",
            expected_result=True,
            log={
                "eventName": "CreateToken",
                "eventSource": "sso.amazonaws.com",
                "eventVersion": "1.08",
                "recipientAccountId": "<organization master account ID>",
                "requestParameters": {
                    "clientId": "...",
                    "clientSecret": "HIDDEN_DUE_TO_SECURITY_REASONS",
                    "deviceCode": "...",
                    "grantType": "urn:ietf:params:oauth:grant-type:device_code",
                },
                "responseElements": {
                    "accessToken": "HIDDEN_DUE_TO_SECURITY_REASONS",
                    "expiresIn": 28800,
                    "idToken": "HIDDEN_DUE_TO_SECURITY_REASONS",
                    "refreshToken": "HIDDEN_DUE_TO_SECURITY_REASONS",
                    "tokenType": "Bearer",
                },
                "sourceIPAddress": "<Attacker source IP>",
                "userAgent": "<Attacker user agent (here: Boto3/1.17.80 Python/3.9.5 Darwin/20.3.0 Botocore/1.20.80)>",
                "userIdentity": {
                    "accountId": "<organization master account ID>",
                    "principalId": "<internal victim user id>",
                    "type": "Unknown",
                    "userName": "<victim display name>",
                },
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.msft import is_external_address, m365_alert_context


@panther_managed
class Microsoft365ExchangeExternalForwarding(Rule):
    default_description = "Detects when a user creates email forwarding rules to external organizations in Microsoft Exchange Online. This can indicate data exfiltration attempts, where an attacker sets up forwarding to collect emails outside the organization. The rule detects both mailbox forwarding (Set-Mailbox) and inbox rules (New-InboxRule).\nThe detection includes: 1. External organization forwarding based on domain comparison 2. Suspicious forwarding patterns like:\n   - Forwarding without keeping a copy\n   - Deleting messages after forwarding\n   - Stopping rule processing after forwarding\n3. Multiple forwarding destinations 4. Various forwarding methods (SMTP, redirect, forward as attachment)\n"
    display_name = "Microsoft Exchange External Forwarding"
    reports = {"MITRE ATT&CK": ["TA0003:T1137.005", "TA0009:T1114.003", "TA0010:T1020"]}
    default_reference = "https://learn.microsoft.com/en-us/microsoft-365/security/office-365-security/outbound-spam-policies-external-email-forwarding?view=o365-worldwide"
    default_severity = Severity.HIGH
    tags = ["Microsoft365", "Exchange", "Data Exfiltration", "Email Security"]
    log_types = [LogType.MICROSOFT365_AUDIT_EXCHANGE]
    id = "Microsoft365.Exchange.External.Forwarding-prototype"
    summary_attributes = ["userid", "parameters", "organizationname"]
    default_runbook = "1. Investigate the forwarding configuration:\n   - Check if the forwarding is legitimate and approved\n   - Verify the destination addresses\n   - Review any suspicious patterns (deletion, no copy kept)\n2. If unauthorized:\n   - Remove the forwarding rule\n   - Check for any data that may have been forwarded\n   - Review the user's recent activity\n3. If authorized:\n   - Document the business justification\n   - Ensure it complies with security policies\n   - Monitor for any changes to the forwarding configuration\n"
    FORWARDING_PARAMETERS = {
        "ForwardingSmtpAddress",
        "ForwardTo",
        "ForwardingAddress",
        "RedirectTo",
        "ForwardAsAttachmentTo",
    }  # Only forward, don't keep copy
    # Delete after forwarding
    # Stop processing other rules
    SUSPICIOUS_PATTERNS = {
        "DeliverToMailboxAndForward": "False",
        "DeleteMessage": "True",
        "StopProcessingRules": "True",
    }

    def rule(self, event):
        """Alert on suspicious or external email forwarding configurations."""
        # Skip non-forwarding related operations
        if event.get("operation") not in ("Set-Mailbox", "New-InboxRule"):
            return False
        # Get organization domains from userid and organizationname
        onmicrosoft_domain = event.get("organizationname", "").lower()
        userid = event.get("userid", "").lower()
        try:
            primary_domain = userid.split("@")[1]
        except (IndexError, AttributeError):
            primary_domain = onmicrosoft_domain if onmicrosoft_domain else None
        if not primary_domain:
            return True  # Alert if we can't determine organization
        # Check each parameter
        for param in event.get("parameters", []):
            param_name = param.get("Name", "")
            param_value = param.get("Value", "")
            # Check for suspicious patterns
            if param_name in self.SUSPICIOUS_PATTERNS and param_value == self.SUSPICIOUS_PATTERNS[param_name]:
                return True
            # Check for external forwarding
            if param_name in self.FORWARDING_PARAMETERS and param_value:
                if is_external_address(param_value, primary_domain, onmicrosoft_domain):
                    return True
        return False

    def title(self, event):
        parameters = event.get("parameters", [])
        forwarding_addresses = []
        suspicious_configs = []
        for param in parameters:
            param_name = param.get("Name", "")
            param_value = param.get("Value", "")
            if param_name in self.FORWARDING_PARAMETERS and param_value:
                # Handle smtp: prefix
                if param_value.lower().startswith("smtp:"):
                    param_value = param_value[5:]
                # Handle multiple addresses
                addresses = param_value.split(";")
                forwarding_addresses.extend(addr.strip() for addr in addresses if addr.strip())
            if param_name in self.SUSPICIOUS_PATTERNS and param_value == self.SUSPICIOUS_PATTERNS[param_name]:
                suspicious_configs.append(f"{param_name}={param_value}")
        to_emails = ", ".join(forwarding_addresses) if forwarding_addresses else "<no-recipient-found>"
        suspicious_str = f" [Suspicious: {', '.join(suspicious_configs)}]" if suspicious_configs else ""
        return f"Microsoft365: External Forwarding Created From [{event.get('userid', '')}] to [{to_emails}]{suspicious_str}"

    def alert_context(self, event):
        return m365_alert_context(event)

    tests = [
        RuleTest(
            name="External Organization Forwarding",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "homer.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:peter.griffin@familyguy.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "False"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.onmicrosoft.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Internal Organization Forwarding",
            expected_result=False,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "marge.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:marge.simpson@simpsons.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "True"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Suspicious Forwarding Pattern",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "New-InboxRule",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "Delete and Forward Rule"},
                    {"Name": "Mailbox", "Value": "homer.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardTo", "Value": "external@example.com"},
                    {"Name": "DeleteMessage", "Value": "True"},
                    {"Name": "StopProcessingRules", "Value": "True"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.onmicrosoft.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Multiple Forwarding Addresses",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "New-InboxRule",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "Multiple Forward Rule"},
                    {"Name": "Mailbox", "Value": "homer.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardTo", "Value": "external1@example.com;external2@example.com;external3@example.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "False"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.onmicrosoft.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Invalid Identity Format",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "Invalid/Path/Format"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:hello@familyguy.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "False"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.onmicrosoft.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Missing Organization Name",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {
                        "Name": "Identity",
                        "Value": "ABC1.prod.outlook.com/Microsoft Exchange Hosted Organizations/simpsons.onmicrosoft.com/homer.simpson",
                    },
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:hello@familyguy.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "False"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.onmicrosoft.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Subdomain Forwarding (Internal)",
            expected_result=False,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "homer.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:bart.simpson@springfield.simpsons.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "True"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Similar Domain Forwarding (External)",
            expected_result=True,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "homer.simpson@simpsons.onmicrosoft.com"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:evil@simpsons2.com"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "True"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.com",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
        RuleTest(
            name="Non-Com TLD Organization",
            expected_result=False,
            log={
                "clientip": "1.2.3.4",
                "creationtime": "2022-12-12 22:19:00",
                "externalaccess": False,
                "id": "111-22-33",
                "objectid": "homer.simpson",
                "operation": "Set-Mailbox",
                "organizationid": "11-aa-bb",
                "organizationname": "simpsons.onmicrosoft.com",
                "originatingserver": "QWERTY (1.2.3.4)",
                "parameters": [
                    {"Name": "Identity", "Value": "homer.simpson@simpsons.org"},
                    {"Name": "ForwardingSmtpAddress", "Value": "smtp:marge.simpson@simpsons.org"},
                    {"Name": "DeliverToMailboxAndForward", "Value": "True"},
                ],
                "recordtype": 1,
                "resultstatus": "True",
                "userid": "homer.simpson@simpsons.org",
                "userkey": "12345",
                "usertype": 2,
                "workload": "Exchange",
            },
        ),
    ]



from panther_core import PantherEvent

from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.aws import aws_cloudtrail_success, aws_rule_context, lookup_aws_account_name


@panther_managed
class AWSEC2MultiInstanceConnect(Rule):
    id = "AWS.EC2.MultiInstanceConnect-prototype"
    display_name = "AWS EC2 Multi Instance Connect"
    log_types = [LogType.AWS_CLOUDTRAIL]
    default_severity = Severity.HIGH
    reports = {"MITRE ATT&CK": ["TA0008:T1021.005"]}
    default_description = "Detect when an attacker pushes an SSH public key to multiple EC2 instances.\n"
    threshold = 2
    default_reference = (
        "https://stratus-red-team.cloud/attack-techniques/AWS/aws.lateral-movement.ec2-instance-connect/\n"
    )
    default_runbook = "Followup with the actor to determine if the SSH key is genuine. Consider using a different SSH key for each instance.\n"
    summary_attributes = ["p_any_actor_ids", "p_any_aws_account_ids", "p_any_aws_instance_ids", "p_any_usernames"]
    tags = ["AWS CloudTrail", "Lateral Movement", "Remote Services", "SSH", "Lateral Movement:Remote Services"]

    def rule(self, event: PantherEvent) -> bool:
        return aws_cloudtrail_success(event) and event.get("eventName") == "SendSSHPublicKey"

    def title(self, event: PantherEvent) -> str:
        actor = event.udm("actor_user")
        account_name = lookup_aws_account_name(event.get("recipientAccountId"))
        return f"{actor} uploaded an SSH Key to multiple instances in {account_name}"

    def dedup(self, event: PantherEvent) -> str:
        # Dedup based on the public SSH key
        return event.deep_get("requestParameters", "sSHPublicKey")

    def alert_context(self, event: PantherEvent) -> dict:
        context = aws_rule_context(event)
        context["instanceId"] = event.deep_get("requestParameters", "instanceId", default="<UNKNOWN EC2 INSTANCE ID>")
        return context

    tests = [
        RuleTest(
            name="Public SSH Key Sent Successfully",
            expected_result=True,
            log={
                "p_event_time": "2025-01-13 19:58:59.000000000",
                "p_log_type": "AWS.CloudTrail",
                "p_parse_time": "2025-01-13 20:05:54.575569351",
                "awsRegion": "us-west-2",
                "eventCategory": "Management",
                "eventID": "d6d05dd2-d03d-4dce-a88c-02b6a567d889",
                "eventName": "SendSSHPublicKey",
                "eventSource": "ec2-instance-connect.amazonaws.com",
                "eventTime": "2025-01-13 19:58:59.000000000",
                "eventType": "AwsApiCall",
                "eventVersion": "1.08",
                "managementEvent": True,
                "readOnly": False,
                "recipientAccountId": "111122223333",
                "requestID": "25eac5d1-cd24-4156-a49b-f2bf3a20ec9d",
                "requestParameters": {
                    "instanceId": "i-abcdef01234567890",
                    "instanceOSUser": "ec2-user",
                    "sSHPublicKey": "ssh-ed25519 XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                },
                "responseElements": {"requestId": "25eac5d1-cd24-4156-a49b-f2bf3a20ec9d", "success": True},
                "sourceIPAddress": "1.2.3.4",
                "tlsDetails": {
                    "cipherSuite": "TLS_AES_128_GCM_SHA256",
                    "clientProvidedHostHeader": "ec2-instance-connect.us-west-2.amazonaws.com",
                    "tlsVersion": "TLSv1.3",
                },
                "userAgent": "stratus-red-team_8b255a24-d33d-4750-bd4b-4007124741df",
                "userIdentity": {
                    "accessKeyId": "SAMPLE_ACCESS_KEY",
                    "accountId": "111122223333",
                    "arn": "arn:aws:sts::111122223333:assumed-role/SampleRole/bobson.dugnutt",
                    "principalId": "SAMPLE_PRINCIPAL_ID:bobson.dugnutt",
                    "sessionContext": {
                        "attributes": {"creationDate": "2025-01-13T19:21:25Z", "mfaAuthenticated": "false"},
                        "sessionIssuer": {
                            "accountId": "111122223333",
                            "arn": "arn:aws:iam::111122223333:role/aws-reserved/sso.amazonaws.com/us-west-2/SampleRole",
                            "principalId": "SAMPLE_PRINCIPAL_ID",
                            "type": "Role",
                            "userName": "SampleRole",
                        },
                        "webIdFederationData": {},
                    },
                    "type": "AssumedRole",
                },
            },
        ),
        RuleTest(
            name="Public SSH Key Not Sent Successfully",
            expected_result=False,
            log={
                "p_event_time": "2025-01-13 20:46:57.000000000",
                "p_log_type": "AWS.CloudTrail",
                "p_parse_time": "2025-01-13 20:55:54.194562429",
                "awsRegion": "us-west-2",
                "errorCode": "AccessDenied",
                "errorMessage": "User: arn:aws:sts::111122223333:assumed-role/SampleRole/bobson.dugnutt is not authorized to perform: ec2-instance-connect:SendSSHPublicKey on resource: arn:aws:ec2:us-west-2:111122223333:instance/i-abcdef01234567890 because no identity-based policy allows the ec2-instance-connect:SendSSHPublicKey action",
                "eventCategory": "Management",
                "eventID": "3ad527a2-8799-4561-8def-963a6dcdbbb5",
                "eventName": "SendSSHPublicKey",
                "eventSource": "ec2-instance-connect.amazonaws.com",
                "eventTime": "2025-01-13 20:46:57.000000000",
                "eventType": "AwsApiCall",
                "eventVersion": "1.08",
                "managementEvent": True,
                "readOnly": False,
                "recipientAccountId": "111122223333",
                "requestID": "cd8f31b6-a30f-4b14-a06a-3b41012228f3",
                "sourceIPAddress": "1.2.3.4",
                "tlsDetails": {
                    "cipherSuite": "TLS_AES_128_GCM_SHA256",
                    "clientProvidedHostHeader": "ec2-instance-connect.us-west-2.amazonaws.com",
                    "tlsVersion": "TLSv1.3",
                },
                "userAgent": "stratus-red-team_95b154bb-bcb2-4582-82f4-b3dbb6cfc1d9",
                "userIdentity": {
                    "accessKeyId": "SAMPLE_ACCESS_KEY",
                    "accountId": "111122223333",
                    "arn": "arn:aws:sts::111122223333:assumed-role/SampleRole/bobson.dugnutt",
                    "principalId": "SAMPLE_PRINCIPAL_ID:bobson.dugnutt",
                    "sessionContext": {
                        "attributes": {"creationDate": "2025-01-13T20:46:53Z", "mfaAuthenticated": "false"},
                        "sessionIssuer": {
                            "accountId": "111122223333",
                            "arn": "arn:aws:iam::111122223333:role/aws-reserved/sso.amazonaws.com/us-west-2/SampleRole",
                            "principalId": "SAMPLE_PRINCIPAL_ID",
                            "type": "Role",
                            "userName": "SampleRole",
                        },
                        "webIdFederationData": {},
                    },
                    "type": "AssumedRole",
                },
            },
        ),
        RuleTest(
            name="Unrelated Event",
            expected_result=False,
            log={
                "eventVersion": "1.05",
                "userIdentity": {
                    "type": "AssumedRole",
                    "principalId": "1111:tester",
                    "arn": "arn:aws:sts::123456789012:assumed-role/tester",
                    "accountId": "123456789012",
                    "accessKeyId": "1",
                    "sessionContext": {
                        "attributes": {"mfaAuthenticated": "true", "creationDate": "2019-01-01T00:00:00Z"},
                        "sessionIssuer": {
                            "type": "Role",
                            "principalId": "1111",
                            "arn": "arn:aws:iam::123456789012:role/tester",
                            "accountId": "123456789012",
                            "userName": "tester",
                        },
                    },
                },
                "eventTime": "2019-01-01T00:00:00Z",
                "eventSource": "ec2.amazonaws.com",
                "eventName": "CreateNetworkAclEntry",
                "awsRegion": "us-west-2",
                "sourceIPAddress": "111.111.111.111",
                "userAgent": "console.ec2.amazonaws.com",
                "requestParameters": {
                    "networkAclId": "acl-1",
                    "ruleNumber": 500,
                    "egress": True,
                    "ruleAction": "allow",
                    "icmpTypeCode": {},
                    "portRange": {},
                    "aclProtocol": "-1",
                    "cidrBlock": "0.0.0.0/0",
                },
                "responseElements": {"requestID": "1", "_return": True},
                "requestID": "1",
                "eventID": "1",
                "eventType": "AwsApiCall",
                "recipientAccountId": "123456789012",
            },
        ),
    ]



"""
Used to define global helpers and variables for CrowdStrike FDR events.
"""


def crowdstrike_detection_alert_context(event):
    """Returns common context for Crowdstrike detections"""
    return {
        "aid": get_crowdstrike_field(event, "aid", default=""),
        "user": get_crowdstrike_field(event, "UserName", default=""),
        "console-link": get_crowdstrike_field(event, "FalconHostLink", default=""),
        "commandline": get_crowdstrike_field(event, "CommandLine", default=""),
        "parentcommandline": get_crowdstrike_field(event, "ParentCommandLine", default=""),
        "filename": get_crowdstrike_field(event, "FileName", default=""),
        "filepath": get_crowdstrike_field(event, "FilePath", default=""),
        "description": get_crowdstrike_field(event, "DetectDescription", default=""),
        "action": get_crowdstrike_field(event, "PatternDispositionDescription", default=""),
    }


def crowdstrike_process_alert_context(event):
    """Returns common process context for Crowdstrike detections"""
    return {
        "aid": get_crowdstrike_field(event, "aid", default=""),
        "CommandLine": get_crowdstrike_field(event, "CommandLine", default=""),
        "TargetProcessId": get_crowdstrike_field(event, "TargetProcessId", default=""),
        "RawProcessId": get_crowdstrike_field(event, "RawProcessId", default=""),
        "ParentBaseFileName": get_crowdstrike_field(event, "ParentBaseFileName", default=""),
        "ParentProcessId": get_crowdstrike_field(event, "ParentProcessId", default=""),
        "ImageFileName": get_crowdstrike_field(event, "ImageFileName", default=""),
        "SHA256Hash": get_crowdstrike_field(event, "SHA256HashData", default=""),
        "platform": get_crowdstrike_field(event, "event_platform", default=""),
    }


def crowdstrike_network_detection_alert_context(event):
    """Returns common network context for Crowdstrike detections"""
    return {
        "LocalAddressIP4": get_crowdstrike_field(event, "LocalAddressIP4", default=""),
        "LocalPort": get_crowdstrike_field(event, "LocalPort", default=""),
        "RemoteAddressIP4": get_crowdstrike_field(event, "RemoteAddressIP4", default=""),
        "RemotePort": get_crowdstrike_field(event, "RemotePort", default=""),
        "Protocol": get_crowdstrike_field(event, "Protocol", default=""),
        "event_simpleName": get_crowdstrike_field(event, "event_simpleName", default=""),
        "aid": get_crowdstrike_field(event, "aid", default=""),
        "ContextProcessId": get_crowdstrike_field(event, "ContextProcessId", default=""),
    }


def filter_crowdstrike_fdr_event_type(event, name: str) -> bool:
    """
    Checks if the event belongs to the Crowdstrike.FDREvent log type
    and the event type is not the name parameter.
    """
    if event.get("p_log_type") != "Crowdstrike.FDREvent":
        return False
    return event.get("fdr_event_type", "") != name


def get_crowdstrike_field(event, field_name, default=None):
    return (
        event.deep_get(field_name)
        or event.deep_get("event", field_name)
        or event.deep_get("unknown_payload", field_name)
        or default
    )



import logging
import os
from dataclasses import dataclass
from fnmatch import fnmatch
from itertools import filterfalse
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast

from ruamel.yaml import YAML
from ruamel.yaml.composer import ComposerError
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError

from pypanther import cli_output
from pypanther.backend.client import BackendError, BackendResponse, ListSchemasParams, Schema, UpdateSchemaParams
from pypanther.backend.client import Client as BackendClient


@dataclass
class SchemaModification:
    # The path of the schema definition file
    filename: str
    # The schema name / identifier, e.g. Custom.SampleSchema
    name: Optional[str]
    # The Backend Client invocation response payload (PutUserSchema endpoint)
    backend_response: Optional[BackendResponse] = None
    # The schema specification in YAML form
    definition: Optional[dict[str, Any]] = None
    # Any error encountered during processing will be stored here
    error: Optional[str] = None
    # Flag to signify whether the schema did exist before
    existed: Optional[bool] = None
    # Flag to signify whether the schema was modified
    modified: Optional[bool] = None
    # Schema Object to be uploaded
    schema: Optional[Schema] = None


@dataclass
class ProcessedFile:
    # The deserialized schema
    yaml: Optional[dict[str, Any]] = None
    # The raw file contents
    raw: str = ""
    # Any error message produced during YAML parsing
    error: Optional[str] = None


class Manager:
    _SCHEMA_NAME_PREFIX = "Custom."
    _SCHEMA_FILE_GLOB_PATTERNS = ("*.yml", "*.yaml")

    def __init__(self, schemas_path: str, verbose: bool, dry_run: bool, backend_client: Optional[BackendClient] = None):
        absolute_path = normalize_path(schemas_path)
        if not absolute_path:
            if verbose:
                print(cli_output.warning("Schemas directory not found. Skipping schemas upload."))
            return

        self._path = absolute_path
        self._backend: Optional[BackendClient] = backend_client
        self._dry_run = dry_run
        self._files: Optional[List[str]] = None
        self._existing_upstream_schemas: Optional[List[Schema]] = None

        self._schemas_to_write = self.prepare()

        # we need to print errors from local files from this early on
        report_summary(absolute_path, self._schemas_to_write, verbose)

    def prepare(self) -> List[SchemaModification]:
        """
        Processes all potential schema files found in the given path.
        """
        if not self.files:
            logging.debug("No files found in path '%s'", self._path)
            return []

        processed_files = self.load_from_yaml(self.files)
        results = []
        for filename, processed_file in processed_files.items():
            # Add results for files that could not be loaded first
            # Skip any files with load errors, we have already included them in the previous loop
            if processed_file.error is not None:
                results.append(
                    SchemaModification(
                        name=None,
                        filename=filename,
                        error=processed_file.error,
                    ),
                )
                continue

            processed_yaml = cast(dict[str, Any], processed_file.yaml)  # type assertion. No-op in runtime
            logging.debug("Processing file %s", filename)

            name, error = self.extract_schema_name(processed_yaml, self._SCHEMA_NAME_PREFIX)
            mod = SchemaModification(filename=filename, name=name, error=error)

            if not mod.error:
                mod.schema = self.generate_schema_object(name, processed_file)
            results.append(mod)
        return results

    def check_upstream(self):
        """
        Check the schemas with the backend. This is in order to report changes during dry-runs and whatnot
        """
        for schema_mod in self.schemas:
            schema = schema_mod.schema

            existing_schema = self.find_schema(schema.name)
            if existing_schema is None:
                schema_mod.existed = False
                schema_mod.modified = False
                continue

            schema_mod.existed = True
            schema_mod.modified = schema_has_changed(existing_schema, schema)

            schema.revision = existing_schema.revision
            # even if the schema has not been changed, we will still do the operation to actually make sure we are
            # synced with the backend. If there has been a change we will get an error due to the revision conflict.
            schema.reference_url = existing_schema.reference_url if schema.reference_url == "" else schema.reference_url
            schema.description = existing_schema.description if schema.description == "" else schema.description

    def apply(self, verbose: bool) -> bool:
        """
        Processes all potential schema changes
        For update-ops it is required to retrieve description, revision number,
        and reference URL from the backend for each schema. More specifically:
        - A matching revision number must be provided when making update requests,
          otherwise validation fails.
        """
        errored = False

        for s in self.schemas:
            if s.error:
                continue
            if s.schema is None:
                raise ValueError("schema is None")
            try:
                s.backend_response = self.do_request(s.schema)
            except BackendError as exc:
                errored = True
                s.error = f"failure to update schema {s.name}: " f"message={exc}"

        if len(self.schemas) > 0:
            report_summary(self._path, self.schemas, verbose)
        return errored

    def do_request(self, s: Schema) -> BackendResponse:
        """
        Update or create a schema based on the processed file contents.

        Note: Even if the schema has not been changed, we will still do the operation to actually make sure we are
        synced with the backend. If there has been a change we will get an error due to the revision conflict.
        """
        if self._backend is None:
            raise ValueError("backend is None")
        logging.debug("updating schema '%s' at revision '%d', using ", s.name, s.revision)

        return self._backend.update_schema(
            params=UpdateSchemaParams(
                name=s.name,
                spec=s.spec,
                revision=s.revision,
                reference_url=s.reference_url,
                description=s.description,
                field_discovery_enabled=s.field_discovery_enabled,
            ),
        )

    @staticmethod
    def generate_schema_object(name: str, processed_file: ProcessedFile) -> Schema:
        if not processed_file.yaml:
            raise ValueError("processed_file.yaml is None")

        return Schema(
            name=name,
            spec=processed_file.raw,
            reference_url=processed_file.yaml.get("referenceURL", ""),
            description=processed_file.yaml.get("description", ""),
            field_discovery_enabled=processed_file.yaml.get("fieldDiscoveryEnabled", True),
            # This will be filled in with the correct value later in case it exists
            revision=0,
            # We don't care about the following at this point. We won't use them during the update
            created_at="",
            updated_at="",
            is_managed=False,
        )

    @property
    def schemas(self) -> List[SchemaModification]:
        if hasattr(self, "_schemas_to_write"):
            return self._schemas_to_write
        return []

    @property
    def files(self) -> List[str]:
        """
        Resolves the list of schema definition files.

        Returns
        -------
            A list of absolute paths to the schema files.

        """
        if self._files is None:
            matching_filenames = discover_files(self._path, self._SCHEMA_FILE_GLOB_PATTERNS)
            self._files = ignore_schema_test_files(matching_filenames)
        return self._files

    @property
    def existing_upstream_schemas(self) -> List[Schema]:
        """
        Retrieves and caches in the instance state the list
        of available user-defined schemas.

        Returns
        -------
             List of user-defined schema records.

        """
        if self._backend is None:
            raise ValueError("backend is None")

        if self._existing_upstream_schemas is None:
            resp = self._backend.list_schemas(ListSchemasParams(is_managed=False))
            if not resp.status_code == 200:
                raise RuntimeError("unable to retrieve custom schemas")
            self._existing_upstream_schemas = resp.data.schemas
        return self._existing_upstream_schemas

    def find_schema(self, name: str) -> Optional[Schema]:
        """
        Find schema by name.

        Returns
        -------
             The decoded YAML schema or None if no matching name is found.

        """
        for schema in self.existing_upstream_schemas:
            if schema.name.casefold() == name.casefold():
                return schema
        return None

    @staticmethod
    def load_from_yaml(files: List[str]) -> Dict[str, ProcessedFile]:
        processed_files = {}
        yaml_parser = YAML(typ="safe")

        for filename in files:
            logging.debug("Loading schema from file %s", filename)
            processed_file = ProcessedFile()
            try:
                with open(filename, encoding="utf-8") as schema_file:
                    raw = schema_file.read()
                processed_file.raw = raw
                processed_file.yaml = yaml_parser.load(raw)
            except (ParserError, ScannerError, ComposerError) as exc:
                processed_file.error = f"invalid YAML: {exc}"
            processed_files[filename] = processed_file
        return processed_files

    @staticmethod
    def extract_schema_name(definition: dict[str, Any], prefix: str) -> Tuple[str, Optional[str]]:
        name = definition.get("schema")

        if name is None:
            return "", "key 'schema' not found"

        if not name.startswith(prefix):
            return (
                "",
                f"'schema' field: value must start" f" with the prefix '{prefix}'",
            )

        if len(name) > 255:
            return "", "'schema' field: value should not exceed 255 characters"

        return name, None


def discover_files(base_path: str, patterns: Tuple[str, ...]) -> List[str]:
    """
    Recursively locates files that match the given glob patterns.

    Args:
    ----
         base_path: the base directory for recursively searching for files
         patterns: a list of glob patterns that the filenames should match

    Returns:
    -------
        A sorted list of absolute paths.

    """
    files = []
    for directory, _, filenames in os.walk(base_path):
        for filename in filenames:
            for pattern in patterns:
                if fnmatch(filename, pattern):
                    files.append(os.path.join(directory, filename))
    return sorted(files)


def ignore_schema_test_files(paths: List[str]) -> List[str]:
    """
    Detect and ignore files that contain schema tests.
    """
    return list(filterfalse(_contains_schema_tests, paths))


def _contains_schema_tests(filename: str) -> bool:
    """
    Check if a file contains YAML document(s) that describe test cases for custom schemas.
    Note that a test case file may contain multiple YAML documents.

    We require that files containing test cases have a specific suffix and extension.
    """
    if not filename.endswith("_tests.yml"):
        return False

    yaml_parser = YAML(typ="safe")

    with open(filename, encoding="utf-8") as stream:
        try:
            yaml_documents: List[Dict[str, Any]] = yaml_parser.load_all(stream)
        except (ParserError, ScannerError, ComposerError):
            return False

        documents = list(yaml_documents)

    if not documents:
        return False

    fields = {x.lower() for x in documents[0].keys()}

    # - "input" and "logtype" are expected to be always present
    # - at least one of "result", "results" fields is required
    return {"input", "logtype", "result"}.issubset(fields) or {
        "input",
        "logtype",
        "results",
    }.issubset(fields)


def normalize_path(path: str) -> Optional[str]:
    """
    Resolve the given path to its absolute form, taking into
    account user home prefix notation.
    """
    absolute_path = Path.resolve(Path.expanduser(Path(path)))
    if not os.path.exists(absolute_path):
        return None
    return str(absolute_path)


def report_summary(base_path: str, results: List[SchemaModification], verbose: bool):
    """
    Translate uploader results to descriptive status messages and prints them.
    Prints only on verbose and on errors.
    """
    for result in sorted(results, key=lambda r: r.filename):
        filename = result.filename.split(base_path)[-1].strip(os.path.sep)
        if result.error:
            print(
                cli_output.failed(
                    f"Failed to update schema from definition" f" in file '{filename}':" f"{result.error}",
                ),
            )
        elif verbose:
            if result.modified:
                print(f"Successfully updated schema '{result.name}' from definition in file '{filename}'")
            else:
                print(f"Successfully created schema '{result.name}' from definition in file '{filename}'")


def schema_has_changed(existing_schema: Schema, new_schema: Schema) -> bool:
    """
    Compare the schema definition in the processed file with the existing schema.
    """
    yaml_parser = YAML(typ="safe")

    return yaml_parser.load(existing_schema.spec) != yaml_parser.load(new_schema.spec)



from pypanther.rules.crowdstrike.event_stream.crowdstrike_admin_role_assigned import (
    CrowdstrikeAdminRoleAssigned as CrowdstrikeAdminRoleAssigned,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_allowlist_removed import (
    CrowdstrikeAllowlistRemoved as CrowdstrikeAllowlistRemoved,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_api_key_created import (
    CrowdstrikeAPIKeyCreated as CrowdstrikeAPIKeyCreated,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_api_key_deleted import (
    CrowdstrikeAPIKeyDeleted as CrowdstrikeAPIKeyDeleted,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_detection_summary import (
    CrowdstrikeEppDetectionSummary as CrowdstrikeEppDetectionSummary,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_ip_allowlist_changed import (
    CrowdstrikeIpAllowlistChanged as CrowdstrikeIpAllowlistChanged,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_new_user_created import (
    CrowdstrikeNewUserCreated as CrowdstrikeNewUserCreated,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_password_change import (
    CrowdstrikeUserPasswordChange as CrowdstrikeUserPasswordChange,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_single_ip_allowlisted import (
    CrowdstrikeSingleIpAllowlisted as CrowdstrikeSingleIpAllowlisted,
)
from pypanther.rules.crowdstrike.event_stream.crowdstrike_user_deleted import (
    CrowdstrikeUserDeleted as CrowdstrikeUserDeleted,
)



from pypanther.rules.push_security.push_security_app_banner_acknowledged import (
    PushSecurityAppBannerAcknowledged as PushSecurityAppBannerAcknowledged,
)
from pypanther.rules.push_security.push_security_authorized_idp_login import (
    PushSecurityAuthorizedIdPLogin as PushSecurityAuthorizedIdPLogin,
)
from pypanther.rules.push_security.push_security_mfa_method_changed import (
    PushSecurityMFAMethodChanged as PushSecurityMFAMethodChanged,
)
from pypanther.rules.push_security.push_security_new_app_detected import (
    PushSecurityNewAppDetected as PushSecurityNewAppDetected,
)
from pypanther.rules.push_security.push_security_new_saas_account_created import (
    PushSecurityNewSaaSAccountCreated as PushSecurityNewSaaSAccountCreated,
)
from pypanther.rules.push_security.push_security_open_security_finding import (
    PushSecurityOpenSecurityFinding as PushSecurityOpenSecurityFinding,
)
from pypanther.rules.push_security.push_security_phishable_mfa_method import (
    PushSecurityPhishableMFAMethod as PushSecurityPhishableMFAMethod,
)
from pypanther.rules.push_security.push_security_phishing_attack import (
    PushSecurityPhishingAttack as PushSecurityPhishingAttack,
)
from pypanther.rules.push_security.push_security_unauthorized_idp_login import (
    PushSecurityUnauthorizedIdPLogin as PushSecurityUnauthorizedIdPLogin,
)



"""Default Configuration values for Panther"""

"""
Here, default values for `panther_config.config` are defined
"""

from ipaddress import ip_network

# A list of public DNS domain names that fall under the administrative domain of
# the Panther installation
ORGANIZATION_DOMAINS = ["example.com"]

AWS_ACCOUNTS = {
    # Add your AWS account IDs/names below:
    "123456789012": "sample-account",
}
DROPBOX_ALLOWED_SHARE_DOMAINS = ORGANIZATION_DOMAINS
DROPBOX_TRUSTED_OWNERSHIP_DOMAINS = ORGANIZATION_DOMAINS
GCP_PRODUCTION_PROJECT_IDS = ["example-production", "example-platform"]
GCP_ORG_ID = "888888888888"
GSUITE_TRUSTED_FORWARDING_DESTINATION_DOMAINS = ORGANIZATION_DOMAINS
GSUITE_TRUSTED_OWNERSHIP_DOMAINS = ORGANIZATION_DOMAINS
MS_EXCHANGE_ALLOWED_FORWARDING_DESTINATION_DOMAINS = ORGANIZATION_DOMAINS
MS_EXCHANGE_ALLOWED_FORWARDING_DESTINATION_EMAILS = ["postmaster@" + ORGANIZATION_DOMAINS[0]]
TELEPORT_ORGANIZATION_DOMAINS = ORGANIZATION_DOMAINS

# Expects a map with a Key 'Tags' that maps to a map of key/value string pairs, or None if no
# tags are present.
# All Panther defined resources meet this requirement.
CDE_TAG_KEY = "environment"
CDE_TAG_VALUE = "pci"


DMZ_NETWORKS = [
    # ip_network("10.1.0.0/24"),
]

DMZ_TAGS = set(
    [
        ("environment", "dmz"),
    ],
)


PCI_NETWORKS = [
    # ip_network("10.0.0.0/24"),
]


# Expects a string in cidr notation (e.g. '10.0.0.0/24') indicating the ip range being checked
# Returns True if any ip in the range is marked as DMZ space.
def is_dmz_cidr(ip_range):
    """This function determines whether a given IP range is within the defined DMZ IP range."""
    return any(ip_network(ip_range).overlaps(dmz_network) for dmz_network in DMZ_NETWORKS)


# Defaults to False to assume something is not a DMZ if it is not tagged
def is_dmz_tags(resource, dmz_tags):
    """This function determines whether a given resource is tagged as existing in a DMZ."""
    if resource["Tags"] is None:
        return False
    for key, value in dmz_tags:
        if resource["Tags"].get(key) == value:
            return True
    return False


# Expects a string in cidr notation (e.g. '10.0.0.0/24') indicating the ip range being checked
# Returns True if any ip in the range is marked as in scope
def is_pci_scope_cidr(ip_range):
    return any(ip_network(ip_range).overlaps(pci_network) for pci_network in PCI_NETWORKS)


# Defaults to True to assume something is in scope if it is not tagged
def in_pci_scope_tags(resource):
    if resource.get("Tags") is None:
        return True
    return resource["Tags"].get(CDE_TAG_KEY) == CDE_TAG_VALUE


# Function variables here so that implementation details of these functions can be changed without
# having to rename the function in all locations its used, or having an outdated name on the actual
# function being used, etc.
IN_PCI_SCOPE = in_pci_scope_tags



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.tines import tines_alert_context


@panther_managed
class TinesSSOSettings(Rule):
    id = "Tines.SSO.Settings-prototype"
    display_name = "Tines SSO Settings"
    log_types = [LogType.TINES_AUDIT]
    tags = ["Tines", "IAM - Credential Security"]
    default_severity = Severity.HIGH
    default_description = "Detects when Tines SSO settings are changed\n"
    default_reference = "https://www.tines.com/docs/admin/single-sign-on"
    summary_attributes = ["user_id", "operation_name", "tenant_id", "request_ip"]
    ACTIONS = ["SsoConfigurationDefaultSet", "SsoConfigurationOidcSet", "SsoConfigurationSamlSet"]

    def rule(self, event):
        action = event.get("operation_name", "<NO_OPERATION_NAME>")
        return action in self.ACTIONS

    def title(self, event):
        action = event.get("operation_name", "<NO_OPERATION_NAME>")
        return f"Tines: [{action}] Setting changed by [{event.deep_get('user_email', default='<NO_USEREMAIL>')}]"

    def alert_context(self, event):
        return tines_alert_context(event)

    def dedup(self, event):
        return f"{event.deep_get('user_id', default='<NO_USERID>')}_{event.deep_get('operation_name', default='<NO_OPERATION>')}"

    tests = [
        RuleTest(
            name="Tines SsoConfigurationSamlSet",
            expected_result=True,
            log={
                "created_at": "2023-05-16 23:26:46",
                "id": 1111111,
                "inputs": {
                    "domainId": "REDACTED",
                    "fingerprint": "REDACTED",
                    "idpCertificate": "REDACTED",
                    "targetUrl": "REDACTED",
                },
                "operation_name": "SsoConfigurationSamlSet",
                "request_ip": "12.12.12.12",
                "request_user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
                "tenant_id": "8888",
                "user_email": "user@company.com",
                "user_id": "17171",
                "user_name": "user at company dot com",
            },
        ),
        RuleTest(
            name="Tines Login",
            expected_result=False,
            log={
                "created_at": "2023-05-17 14:45:19",
                "id": 7888888,
                "operation_name": "Login",
                "request_ip": "12.12.12.12",
                "request_user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
                "tenant_id": "8888",
                "user_email": "user@company.com",
                "user_id": "17171",
                "user_name": "user at company dot com",
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed


@panther_managed
class GSuiteSuspiciousLogins(Rule):
    id = "GSuite.SuspiciousLogins-prototype"
    display_name = "Suspicious GSuite Login"
    log_types = [LogType.GSUITE_ACTIVITY_EVENT]
    tags = ["GSuite"]
    default_severity = Severity.MEDIUM
    default_description = "GSuite reported a suspicious login for this user.\n"
    default_reference = "https://support.google.com/a/answer/7102416?hl=en"
    default_runbook = "Checkout the details of the login and verify this behavior with the user to ensure the account wasn't compromised.\n"
    summary_attributes = ["actor:email"]
    SUSPICIOUS_LOGIN_TYPES = {"suspicious_login", "suspicious_login_less_secure_app", "suspicious_programmatic_login"}

    def rule(self, event):
        if event.deep_get("id", "applicationName") != "login":
            return False
        return bool(event.get("name") in self.SUSPICIOUS_LOGIN_TYPES)

    def title(self, event):
        user = event.deep_get("parameters", "affected_email_address")
        if not user:
            user = "<UNKNOWN_USER>"
        return f"A suspicious login was reported for user [{user}]"

    tests = [
        RuleTest(
            name="Normal Login Event",
            expected_result=False,
            log={
                "id": {"applicationName": "login"},
                "kind": "admin#reports#activity",
                "type": "account_warning",
                "name": "login_success",
                "parameters": {"affected_email_address": "bobert@ext.runpanther.io"},
            },
        ),
        RuleTest(
            name="Account Warning Not For Suspicious Login",
            expected_result=False,
            log={
                "id": {"applicationName": "login"},
                "kind": "admin#reports#activity",
                "type": "account_warning",
                "name": "account_disabled_spamming",
                "parameters": {"affected_email_address": "bobert@ext.runpanther.io"},
            },
        ),
        RuleTest(
            name="Account Warning For Suspicious Login",
            expected_result=True,
            log={
                "id": {"applicationName": "login"},
                "kind": "admin#reports#activity",
                "type": "account_warning",
                "name": "suspicious_login",
                "parameters": {"affected_email_address": "bobert@ext.runpanther.io"},
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.msft import azure_rule_context, azure_success, get_target_name


@panther_managed
class AzureAuditRoleChangedPIM(Rule):
    id = "Azure.Audit.RoleChangedPIM-prototype"
    display_name = "Azure Role Changed PIM"
    log_types = [LogType.AZURE_AUDIT]
    default_severity = Severity.MEDIUM
    dedup_period_minutes = 60
    default_description = "This detection looks for a change in member's PIM roles in EntraID\n"
    reports = {"MITRE ATT&CK": ["TA0042:T1586"]}
    default_runbook = "Verify if the role change was authorized and review the affected user. If unauthorized, revert the role change, notify relevant teams,\n"
    default_reference = "https://learn.microsoft.com/en-us/entra/identity/authentication/overview-authentication"
    summary_attributes = ["properties:ServicePrincipalName", "properties:UserPrincipalName", "properties:ipAddress"]

    def rule(self, event):
        operation = event.get("operationName", default="")
        if azure_success(event) and "Add member to role in PIM completed" in operation:
            return True
        return False

    def title(self, event):
        operation_name = event.get("operationName", default="<UNKNOWN OPERATION>")
        actor_name = event.deep_get("properties", "initiatedBy", "user", "userPrincipalName", default="<UNKNOWN USER>")
        target_name = get_target_name(event)
        role = event.deep_walk(
            "properties",
            "targetResources",
            "displayName",
            return_val="first",
            default="<UNKNOWN_ROLE>",
        )
        return f"{actor_name} added {target_name} as {role} successfully with {operation_name}"

    def alert_context(self, event):
        return azure_rule_context(event)

    tests = [
        RuleTest(
            name="Successfully added PIM role",
            expected_result=True,
            log={
                "category": "AuditLogs",
                "correlationId": "1234155",
                "durationMs": 0,
                "identity": "Ju Cho",
                "Level": 4,
                "operationName": "Add member to role in PIM completed (permanent)",
                "operationVersion": "1.0",
                "properties": {
                    "activityDateTime": "2024-12-16 16:32:16.087554000",
                    "activityDisplayName": "Add member to role in PIM completed (permanent)",
                    "additionalDetails": [
                        {"key": "RoleDefinitionOriginId", "value": "123451235"},
                        {"key": "RoleDefinitionOriginType", "value": "BuiltInRole"},
                        {"key": "TemplateId", "value": "123412351"},
                        {"key": "StartTime", "value": "2024-12-16T16:32:15.8441686Z"},
                        {"key": "Justification", "value": "test assign"},
                        {"key": "oid", "value": "12351534"},
                        {"key": "tid", "value": "345667733"},
                        {"key": "wids", "value": "234523454"},
                        {"key": "ipaddr", "value": "1.2.3.4"},
                        {"key": "RequestId", "value": "651346123452"},
                    ],
                    "category": "RoleManagement",
                    "correlationId": "12345",
                    "id": "PIM_123415",
                    "initiatedBy": {
                        "user": {
                            "displayName": "Ju Cho",
                            "id": "12345",
                            "roles": [],
                            "userPrincipalName": "Radahn@Starscourge.onmicrosoft.com",
                        },
                    },
                    "loggedByService": "PIM",
                    "operationType": "Update",
                    "result": "success",
                    "resultReason": "test assign",
                    "targetResources": [
                        {
                            "administrativeUnits": [],
                            "displayName": "Application Administrator",
                            "id": "12345",
                            "modifiedProperties": [
                                {"displayName": "RoleDefinitionOriginId", "newValue": '"12345"', "oldValue": '""'},
                                {
                                    "displayName": "RoleDefinitionOriginType",
                                    "newValue": '"BuiltInRole"',
                                    "oldValue": '""',
                                },
                                {"displayName": "TemplateId", "newValue": '"12345"', "oldValue": '""'},
                            ],
                            "type": "Role",
                        },
                        {"administrativeUnits": [], "id": "12345", "type": "Request"},
                        {"administrativeUnits": [], "displayName": "Malenia", "id": "12345", "type": "User"},
                        {"administrativeUnits": [], "displayName": "Panther", "id": "12345", "type": "Directory"},
                        {"administrativeUnits": [], "id": "12345", "type": "Other"},
                    ],
                },
                "resourceId": "/tenants/12345/providers/Microsoft.aadiam",
                "resultSignature": "None",
                "tenantId": "12345",
                "time": "2024-12-16 16:32:16.087554000",
            },
        ),
        RuleTest(
            name="requested adding PIM role",
            expected_result=False,
            log={
                "category": "AuditLogs",
                "correlationId": "1234155",
                "durationMs": 0,
                "identity": "Ju Cho",
                "Level": 4,
                "operationName": "Add member to role in PIM requested (permanent)",
                "operationVersion": "1.0",
                "properties": {
                    "activityDateTime": "2024-12-16 16:32:16.087554000",
                    "activityDisplayName": "Add member to role in PIM requested (permanent)",
                    "additionalDetails": [
                        {"key": "RoleDefinitionOriginId", "value": "123451235"},
                        {"key": "RoleDefinitionOriginType", "value": "BuiltInRole"},
                        {"key": "TemplateId", "value": "123412351"},
                        {"key": "StartTime", "value": "2024-12-16T16:32:15.8441686Z"},
                        {"key": "Justification", "value": "test assign"},
                        {"key": "oid", "value": "12351534"},
                        {"key": "tid", "value": "345667733"},
                        {"key": "wids", "value": "234523454"},
                        {"key": "ipaddr", "value": "1.2.3.4"},
                        {"key": "RequestId", "value": "651346123452"},
                    ],
                    "category": "RoleManagement",
                    "correlationId": "12345",
                    "id": "PIM_123415",
                    "initiatedBy": {
                        "user": {
                            "displayName": "Ju Cho",
                            "id": "12345",
                            "roles": [],
                            "userPrincipalName": "Radahn@Starscourge.onmicrosoft.com",
                        },
                    },
                    "loggedByService": "PIM",
                    "operationType": "Update",
                    "result": "success",
                    "resultReason": "test assign",
                    "targetResources": [
                        {
                            "administrativeUnits": [],
                            "displayName": "Application Administrator",
                            "id": "12345",
                            "modifiedProperties": [
                                {"displayName": "RoleDefinitionOriginId", "newValue": '"12345"', "oldValue": '""'},
                                {
                                    "displayName": "RoleDefinitionOriginType",
                                    "newValue": '"BuiltInRole"',
                                    "oldValue": '""',
                                },
                                {"displayName": "TemplateId", "newValue": '"12345"', "oldValue": '""'},
                            ],
                            "type": "Role",
                        },
                        {"administrativeUnits": [], "id": "12345", "type": "Request"},
                        {"administrativeUnits": [], "displayName": "Malenia", "id": "12345", "type": "User"},
                        {"administrativeUnits": [], "displayName": "Panther", "id": "12345", "type": "Directory"},
                        {"administrativeUnits": [], "id": "12345", "type": "Other"},
                    ],
                },
                "resourceId": "/tenants/12345/providers/Microsoft.aadiam",
                "resultSignature": "None",
                "tenantId": "12345",
                "time": "2024-12-16 16:32:16.087554000",
            },
        ),
        RuleTest(
            name="Add member to role (Non PIM)",
            expected_result=False,
            log={
                "category": "AuditLogs",
                "correlationId": "1234155",
                "durationMs": 0,
                "identity": "Ju Cho",
                "Level": 4,
                "operationName": "Add member to role",
                "operationVersion": "1.0",
                "properties": {
                    "activityDateTime": "2024-12-16 16:32:16.087554000",
                    "activityDisplayName": "Add member to role",
                    "additionalDetails": [
                        {"key": "RoleDefinitionOriginId", "value": "123451235"},
                        {"key": "RoleDefinitionOriginType", "value": "BuiltInRole"},
                        {"key": "TemplateId", "value": "123412351"},
                        {"key": "StartTime", "value": "2024-12-16T16:32:15.8441686Z"},
                        {"key": "Justification", "value": "test assign"},
                        {"key": "oid", "value": "12351534"},
                        {"key": "tid", "value": "345667733"},
                        {"key": "wids", "value": "234523454"},
                        {"key": "ipaddr", "value": "1.2.3.4"},
                        {"key": "RequestId", "value": "651346123452"},
                    ],
                    "category": "RoleManagement",
                    "correlationId": "12345",
                    "id": "PIM_123415",
                    "initiatedBy": {
                        "user": {
                            "displayName": "Ju Cho",
                            "id": "12345",
                            "roles": [],
                            "userPrincipalName": "Radahn@Starscourge.onmicrosoft.com",
                        },
                    },
                    "loggedByService": "PIM",
                    "operationType": "Update",
                    "result": "success",
                    "resultReason": "test assign",
                    "targetResources": [
                        {
                            "administrativeUnits": [],
                            "displayName": "Application Administrator",
                            "id": "12345",
                            "modifiedProperties": [
                                {"displayName": "RoleDefinitionOriginId", "newValue": '"12345"', "oldValue": '""'},
                                {
                                    "displayName": "RoleDefinitionOriginType",
                                    "newValue": '"BuiltInRole"',
                                    "oldValue": '""',
                                },
                                {"displayName": "TemplateId", "newValue": '"123415"', "oldValue": '""'},
                            ],
                            "type": "Role",
                        },
                        {"administrativeUnits": [], "id": "12345", "type": "Request"},
                        {"administrativeUnits": [], "displayName": "Malenia", "id": "12345", "type": "User"},
                        {"administrativeUnits": [], "displayName": "Panther", "id": "12345", "type": "Directory"},
                        {"administrativeUnits": [], "id": "12345", "type": "Other"},
                    ],
                },
                "resourceId": "/tenants/12345/providers/Microsoft.aadiam",
                "resultSignature": "None",
                "tenantId": "12345",
                "time": "2024-12-16 16:32:16.087554000",
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.zscaler import zia_alert_context, zia_success


@panther_managed
class ZIAAdditionalCloudRoles(Rule):
    id = "ZIA.Additional.Cloud.Roles-prototype"
    default_description = "This rule detects when an additional cloud role was created."
    display_name = "ZIA Additional Cloud Roles"
    default_runbook = (
        "Verify that this change was planned. If not, revert the change and ensure this doesn't happen again."
    )
    default_reference = "https://help.zscaler.com/zia/about-role-management"
    default_severity = Severity.MEDIUM
    reports = {"MITRE ATT&CK": ["TA0003:T1098.003", "TA0004:T1098.003"]}
    log_types = [LogType.ZSCALER_ZIA_ADMIN_AUDIT_LOG]

    def rule(self, event):
        if not zia_success(event):
            return False
        event_data = event.get("event", {})
        return (
            event_data.get("action", "ACTION_NOT_FOUND") == "CREATE"
            and event_data.get("category", "CATEGORY_NOT_FOUND") == "ROLE_MANAGEMENT"
        )

    def title(self, event):
        return f"[Zscaler.ZIA]: Additional cloud role was created by admin with id [{event.deep_get('event', 'adminid', default='<ADMIN_ID_NOT_FOUND>')}]"

    def alert_context(self, event):
        return zia_alert_context(event)

    tests = [
        RuleTest(
            name="Administration > User Management > Add User, assign to Service Admin group",
            expected_result=False,
            log={
                "event": {
                    "action": "UPDATE",
                    "adminid": "admin@16991311.zscalerbeta.net",
                    "auditlogtype": "ZIA",
                    "category": "USER_MANAGEMENT",
                    "clientip": "123.123.123.123",
                    "errorcode": "None",
                    "interface": "UI",
                    "postaction": {
                        "adminUser": False,
                        "department": {
                            "id": 19752838,
                            "isDeleted": False,
                            "isForUnauthenticatedUser": False,
                            "isNonEditable": False,
                            "name": "test",
                        },
                        "email": "johndoe@dev-company.com",
                        "groups": [
                            {"id": 16991312, "isNonEditable": True, "name": "Service Admin"},
                            {"id": 19631231, "isNonEditable": False, "name": "test"},
                        ],
                        "id": 19752821,
                        "isNonEditable": False,
                        "miscflags": 0,
                        "name": "johndoe",
                        "systemDefinedGroups": [],
                    },
                    "preaction": {
                        "adminUser": False,
                        "authType": "SAFECHANNEL_DIR",
                        "department": {
                            "id": 19752838,
                            "isDeleted": False,
                            "isForUnauthenticatedUser": False,
                            "isNonEditable": False,
                            "name": "test",
                        },
                        "email": "johndoe@dev-company.com",
                        "groups": [{"id": 19631231, "isNonEditable": False, "name": "test"}],
                        "id": 19752821,
                        "miscflags": 268435456,
                        "name": "johndoe",
                    },
                    "recordid": "325",
                    "resource": "johndoe",
                    "result": "SUCCESS",
                    "subcategory": "USER",
                    "time": "2024-10-22 22:02:29.000000000",
                },
                "sourcetype": "zscalernss-audit",
            },
        ),
        RuleTest(
            name="Administration > Role Management > Add Administrator Role, all permissions",
            expected_result=True,
            log={
                "event": {
                    "action": "CREATE",
                    "adminid": "admin@16991311.zscalerbeta.net",
                    "auditlogtype": "ZIA",
                    "category": "ROLE_MANAGEMENT",
                    "clientip": "123.123.123.123",
                    "errorcode": "None",
                    "interface": "UI",
                    "postaction": {
                        "adminAcctAccess": "READ_WRITE",
                        "alertingAccess": "READ_WRITE",
                        "analysisAccess": "READ_ONLY",
                        "dashboardAccess": "READ_WRITE",
                        "deviceInfoAccess": "READ_ONLY",
                        "id": 32780,
                        "logsLimit": "Unrestricted",
                        "name": "mega admin",
                        "permissions": [
                            "ADVANCED_SETTINGS",
                            "COMPLY",
                            "FIREWALL_DNS",
                            "NSS_CONFIGURATION",
                            "SECURE",
                            "SSL_POLICY",
                            "VZEN_CONFIGURATION",
                            "PARTNER_INTEGRATION",
                            "REMOTE_ASSISTANCE_MANAGEMENT",
                            "LOCATIONS",
                            "VPN_CREDENTIALS",
                            "HOSTED_PAC_FILES",
                            "EZ_AGENT_CONFIGURATIONS",
                            "SECURE_AGENT_NOTIFICATIONS",
                            "PROXY_GATEWAY",
                            "STATIC_IPS",
                            "GRE_TUNNELS",
                            "SUBCLOUDS",
                            "AUTHENTICATION_SETTINGS",
                            "USER_MANAGEMENT",
                            "IDENTITY_PROXY_SETTINGS",
                            "APIKEY_MANAGEMENT",
                            "POLICY_RESOURCE_MANAGEMENT",
                            "CLIENT_CONNECTOR_PORTAL",
                            "CUSTOM_URL_CAT",
                            "OVERRIDE_EXISTING_CAT",
                            "TENANT_PROFILE_MANAGEMENT",
                        ],
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportAccess": "READ_WRITE",
                        "reportTimeDuration": -1,
                        "roleType": "EXEC_INSIGHT_AND_ORG_ADMIN",
                        "usernameAccess": "READ_ONLY",
                    },
                    "preaction": {
                        "adminAcctAccess": "READ_WRITE",
                        "alertingAccess": "READ_WRITE",
                        "analysisAccess": "READ_ONLY",
                        "dashboardAccess": "READ_WRITE",
                        "deviceInfoAccess": "READ_ONLY",
                        "id": 0,
                        "isAuditor": False,
                        "isNonEditable": False,
                        "logsLimit": "Unrestricted",
                        "name": "mega admin",
                        "permissions": [
                            "ADVANCED_SETTINGS",
                            "COMPLY",
                            "FIREWALL_DNS",
                            "NSS_CONFIGURATION",
                            "SECURE",
                            "SSL_POLICY",
                            "VZEN_CONFIGURATION",
                            "PARTNER_INTEGRATION",
                            "REMOTE_ASSISTANCE_MANAGEMENT",
                            "LOCATIONS",
                            "VPN_CREDENTIALS",
                            "HOSTED_PAC_FILES",
                            "EZ_AGENT_CONFIGURATIONS",
                            "SECURE_AGENT_NOTIFICATIONS",
                            "PROXY_GATEWAY",
                            "STATIC_IPS",
                            "GRE_TUNNELS",
                            "SUBCLOUDS",
                            "AUTHENTICATION_SETTINGS",
                            "USER_MANAGEMENT",
                            "IDENTITY_PROXY_SETTINGS",
                            "APIKEY_MANAGEMENT",
                            "POLICY_RESOURCE_MANAGEMENT",
                            "CLIENT_CONNECTOR_PORTAL",
                            "CUSTOM_URL_CAT",
                            "OVERRIDE_EXISTING_CAT",
                            "TENANT_PROFILE_MANAGEMENT",
                        ],
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportAccess": "READ_WRITE",
                        "reportTimeDuration": -1,
                        "roleType": "EXEC_INSIGHT_AND_ORG_ADMIN",
                        "usernameAccess": "READ_ONLY",
                    },
                    "recordid": "341",
                    "resource": "mega admin",
                    "result": "SUCCESS",
                    "subcategory": "ADMINISTRATOR_ROLE",
                    "time": "2024-10-22 22:19:57.000000000",
                },
                "sourcetype": "zscalernss-audit",
            },
        ),
        RuleTest(
            name="Administration > Role Management > Add SD-WAN Partner API Role, all permissions",
            expected_result=True,
            log={
                "event": {
                    "action": "CREATE",
                    "adminid": "admin@16991311.zscalerbeta.net",
                    "auditlogtype": "ZIA",
                    "category": "ROLE_MANAGEMENT",
                    "clientip": "123.123.123.123",
                    "errorcode": "None",
                    "interface": "UI",
                    "postaction": {
                        "adminAcctAccess": "NONE",
                        "alertingAccess": "READ_ONLY",
                        "analysisAccess": "NONE",
                        "dashboardAccess": "NONE",
                        "deviceInfoAccess": "NONE",
                        "id": 32781,
                        "name": "wanny",
                        "permissions": ["LOCATIONS", "VPN_CREDENTIALS", "STATIC_IPS", "GRE_TUNNELS"],
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportAccess": "NONE",
                        "reportTimeDuration": -1,
                        "roleType": "SDWAN",
                        "usernameAccess": "NONE",
                    },
                    "preaction": {
                        "id": 0,
                        "name": "wanny",
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportTimeDuration": -1,
                        "roleType": "SDWAN",
                    },
                    "recordid": "343",
                    "resource": "wanny",
                    "result": "SUCCESS",
                    "subcategory": "ADMINISTRATOR_ROLE",
                    "time": "2024-10-22 22:31:46.000000000",
                },
                "sourcetype": "zscalernss-audit",
            },
        ),
        RuleTest(
            name="Administration > Role Management > Add API Role, all permissions",
            expected_result=True,
            log={
                "event": {
                    "action": "CREATE",
                    "adminid": "admin@16991311.zscalerbeta.net",
                    "auditlogtype": "ZIA",
                    "category": "ROLE_MANAGEMENT",
                    "clientip": "123.123.123.123",
                    "errorcode": "None",
                    "interface": "UI",
                    "postaction": {
                        "adminAcctAccess": "READ_WRITE",
                        "alertingAccess": "NONE",
                        "analysisAccess": "NONE",
                        "dashboardAccess": "NONE",
                        "deviceInfoAccess": "NONE",
                        "id": 32782,
                        "logsLimit": "Unrestricted",
                        "name": "bad API",
                        "permissions": [
                            "ADVANCED_SETTINGS",
                            "COMPLY",
                            "FIREWALL_DNS",
                            "SECURE",
                            "SSL_POLICY",
                            "LOCATIONS",
                            "VPN_CREDENTIALS",
                            "STATIC_IPS",
                            "GRE_TUNNELS",
                            "USER_MANAGEMENT",
                            "POLICY_RESOURCE_MANAGEMENT",
                            "CUSTOM_URL_CAT",
                            "OVERRIDE_EXISTING_CAT",
                        ],
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportAccess": "NONE",
                        "reportTimeDuration": -1,
                        "roleType": "PUBLIC_API",
                        "usernameAccess": "NONE",
                    },
                    "preaction": {
                        "adminAcctAccess": "READ_WRITE",
                        "id": 0,
                        "isAuditor": False,
                        "isNonEditable": False,
                        "logsLimit": "Unrestricted",
                        "name": "bad API",
                        "permissions": [
                            "ADVANCED_SETTINGS",
                            "COMPLY",
                            "FIREWALL_DNS",
                            "SECURE",
                            "SSL_POLICY",
                            "LOCATIONS",
                            "VPN_CREDENTIALS",
                            "STATIC_IPS",
                            "GRE_TUNNELS",
                            "USER_MANAGEMENT",
                            "POLICY_RESOURCE_MANAGEMENT",
                            "CUSTOM_URL_CAT",
                            "OVERRIDE_EXISTING_CAT",
                        ],
                        "policyAccess": "READ_WRITE",
                        "rank": 7,
                        "reportTimeDuration": -1,
                        "roleType": "PUBLIC_API",
                    },
                    "recordid": "344",
                    "resource": "bad API",
                    "result": "SUCCESS",
                    "subcategory": "ADMINISTRATOR_ROLE",
                    "time": "2024-10-22 22:34:34.000000000",
                },
                "sourcetype": "zscalernss-audit",
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.duo import duo_alert_context


@panther_managed
class DuoAdminMFARestrictionsUpdated(Rule):
    default_description = "Detects changes to allowed MFA factors administrators can use to log into the admin panel."
    display_name = "Duo Admin MFA Restrictions Updated"
    default_reference = "https://duo.com/docs/essentials-overview"
    default_severity = Severity.MEDIUM
    log_types = [LogType.DUO_ADMINISTRATOR]
    id = "Duo.Admin.MFA.Restrictions.Updated-prototype"

    def rule(self, event):
        return event.get("action") == "update_admin_factor_restrictions"

    def title(self, event):
        return f"Duo Admin MFA Restrictions Updated by [{event.get('username', '<user_not_found>')}]"

    def alert_context(self, event):
        return duo_alert_context(event)

    tests = [
        RuleTest(
            name="Admin MFA Update Event",
            expected_result=True,
            log={
                "action": "update_admin_factor_restrictions",
                "description": '{"allowed_factors": "Duo mobile passcodes, Hardware tokens, Duo push, Yubikey aes"}',
                "isotimestamp": "2022-02-21 21:48:06",
                "timestamp": "2022-02-21 21:48:06",
                "username": "Homer Simpson",
            },
        ),
        RuleTest(
            name="Login Event",
            expected_result=False,
            log={
                "action": "admin_login",
                "description": '{"ip_address": "1.2.3.4", "device": "123-456-789", "factor": "sms", "saml_idp": "OneLogin", "primary_auth_method": "Single Sign-On"}',
                "isotimestamp": "2021-06-30 19:45:37",
                "timestamp": "2021-06-30 19:45:37",
                "username": "Homer Simpson",
            },
        ),
    ]



from __future__ import annotations

from enum import Enum
from functools import total_ordering


@total_ordering
class Severity(str, Enum):
    INFO = "INFO"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"

    def __lt__(self, other):
        return Severity.as_int(self.value) < Severity.as_int(other.value)

    def _increment(self, amt: int = 1) -> Severity:
        """Increase or decrease the severity level by a given amount."""
        # __members__ returns the severities an OrderedDict in the same ordering as the severities
        #    are defined in the class
        severities = list(Severity.__members__.keys())
        current_idx = severities.index(self.name)
        # Change index by ammount, clamping between 0 and 4
        new_idx = min(max(0, current_idx + amt), len(severities) - 1)
        return Severity(severities[new_idx])

    def upgrade(self):
        """
        Increase the severity level by 1. i.e. MEDIUM -> HIGH.
        Cannot return higher than CRITICAL.
        """
        return self._increment(1)

    def downgrade(self):
        """
        Decrease the severity level by 1. i.e. MEDIUM -> LOW.
        Cannot return lower than INFO.
        """
        return self._increment(-1)

    @staticmethod
    def as_int(value: Severity) -> int:
        val = value.upper()
        if val == Severity.INFO:
            return 0
        if val == Severity.LOW:
            return 1
        if val == Severity.MEDIUM:
            return 2
        if val == Severity.HIGH:
            return 3
        if val == Severity.CRITICAL:
            return 4
        raise ValueError(f"Unknown severity: {value}")

    def __str__(self) -> str:
        """Returns a string representation of the class' value."""
        return self.value


# Used to check dynamic severity output
SEVERITY_DEFAULT = "DEFAULT"
SEVERITY_TYPES = [str(sev) for sev in Severity]



from pypanther.rules.aws_s3.aws_s3_access_error import AWSS3ServerAccessError as AWSS3ServerAccessError
from pypanther.rules.aws_s3.aws_s3_access_ip_allowlist import (
    AWSS3ServerAccessIPWhitelist as AWSS3ServerAccessIPWhitelist,
)
from pypanther.rules.aws_s3.aws_s3_insecure_access import AWSS3ServerAccessInsecure as AWSS3ServerAccessInsecure
from pypanther.rules.aws_s3.aws_s3_unauthenticated_access import (
    AWSS3ServerAccessUnauthenticated as AWSS3ServerAccessUnauthenticated,
)
from pypanther.rules.aws_s3.aws_s3_unknown_requester_get_object import (
    AWSS3ServerAccessUnknownRequester as AWSS3ServerAccessUnknownRequester,
)



import argparse
import base64
import importlib
import json
import logging
import os
import tempfile
import time
import zipfile
from dataclasses import asdict
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Optional, Tuple, TypedDict

import requests

from pypanther import cli_output, display, schemas, testing
from pypanther.backend.client import (
    BackendError,
    BulkUploadDetectionsError,
    BulkUploadDetectionsParams,
    BulkUploadDetectionsResults,
    BulkUploadDetectionsStatusParams,
    BulkUploadPresignedURLParams,
)
from pypanther.backend.client import Client as BackendClient
from pypanther.backend.util import convert_unicode
from pypanther.import_main import NoMainModuleError, import_main
from pypanther.registry import registered_rules

INDENT = " " * 2
IGNORE_FOLDERS = [
    ".mypy_cache",
    "pypanther",
    "panther_analysis",
    ".git",
    "__pycache__",
    "tests",
    ".venv",
]

UPLOAD_RESULT_SUCCESS = "UPLOAD_SUCCEEDED"
UPLOAD_RESULT_FAILURE = "UPLOAD_FAILED"
UPLOAD_RESULT_TESTS_FAILED = "TESTS_FAILED"

# There is no hard limit on the BE size, but setting this to 10MB for now to avoid customers uploading by mistake large files
UPLOAD_SIZE_LIMIT_MB = 10
UPLOAD_SIZE_LIMIT_BYTES = UPLOAD_SIZE_LIMIT_MB * 1024 * 1024


class ChangesSummary(TypedDict):
    # rules
    new_rule_ids: list[str]
    delete_rule_ids: list[str]
    modify_rule_ids: list[str]
    total_rule_ids: list[str]
    # schemas
    new_schema_names: list[str]
    modified_schema_names: list[str]
    existed_schema_names: list[str]


def run(backend: BackendClient, args: argparse.Namespace) -> Tuple[int, str]:
    print(
        cli_output.warning(
            "WARNING: pypanther is in beta and is subject to breaking changes before general availability",
        ),
    )
    try:
        import_main(os.getcwd(), "main")
    except NoMainModuleError:
        logging.error("No main.py found. Are you running this command from the root of your pypanther project?")  # noqa: TRY400
        return 1, ""

    test_results = testing.TestResults()
    if not args.skip_tests:
        test_results = testing.run_tests(args)
        if test_results.had_failed_tests():
            if args.output == display.OUTPUT_TYPE_JSON:
                output = get_upload_output_as_dict(
                    None,
                    test_results,
                    [],
                    args.verbose,
                    args.skip_tests,
                    UPLOAD_RESULT_TESTS_FAILED,
                    None,
                )
                print(json.dumps(output, indent=display.JSON_INDENT_LEVEL))
            return 1, ""

        if args.output == display.OUTPUT_TYPE_TEXT:
            print()  # new line to separate test output from upload output

    if args.verbose and args.output == display.OUTPUT_TYPE_TEXT:
        print_registered_rules()

    pypanther_version = importlib.metadata.version("pypanther")

    with tempfile.NamedTemporaryFile() as tmp:
        zip_info = zip_contents(tmp)
        zip_size = Path.stat(Path(tmp.name)).st_size

        if zip_size > UPLOAD_SIZE_LIMIT_BYTES:
            return 1, (
                "Zip file is too big. Reduce number of files or contents of files before "
                f"uploading again (Limit: {UPLOAD_SIZE_LIMIT_MB} MB, Zip size: {zip_size / (1024 * 1024):0.3f} MB)."
            )

        if args.verbose and args.output == display.OUTPUT_TYPE_TEXT:
            print_included_files(zip_info)

        try:
            session_id = upload_zip(
                backend=backend,
                archive=tmp.name,
                verbose=args.verbose,
                output_type=args.output,
                pypanther_version=pypanther_version,
            )
        except BackendError as be_err:
            multi_err = BulkUploadDetectionsError.from_json(convert_unicode(be_err))
            if args.output == display.OUTPUT_TYPE_TEXT:
                print_upload_detection_error(multi_err)
            elif args.output == display.OUTPUT_TYPE_JSON:
                output = get_failed_upload_as_dict(
                    multi_err,
                    test_results,
                    zip_info,
                    args.verbose,
                    args.skip_tests,
                    UPLOAD_RESULT_FAILURE,
                )
                print(json.dumps(output, indent=display.JSON_INDENT_LEVEL))
            return 1, ""

    changes_summary = ChangesSummary(
        new_rule_ids=[],
        delete_rule_ids=[],
        modify_rule_ids=[],
        total_rule_ids=[],
        new_schema_names=[],
        modified_schema_names=[],
        existed_schema_names=[],
    )

    # Prepare schemas first
    manager = schemas.Manager(args.schemas_path, verbose=args.verbose, dry_run=args.dry_run, backend_client=backend)
    manager.check_upstream()
    for schema_to_be_written in manager.schemas:
        if schema_to_be_written.error:  # stop if there's a single error. It's already been printed
            return 1, ""
        if not schema_to_be_written.name:
            raise ValueError("Schema name is required")
        if schema_to_be_written.modified:
            changes_summary["modified_schema_names"].append(schema_to_be_written.name)
        elif schema_to_be_written.existed:
            changes_summary["existed_schema_names"].append(schema_to_be_written.name)
        else:
            changes_summary["new_schema_names"].append(schema_to_be_written.name)

    try:
        if not args.skip_summary or args.dry_run:
            rules_changes_summary = dry_run_rule_upload(
                backend=backend,
                session_id=session_id,
                verbose=args.verbose,
                output_type=args.output,
            )
            changes_summary["new_rule_ids"] = rules_changes_summary["new_rule_ids"]
            changes_summary["delete_rule_ids"] = rules_changes_summary["delete_rule_ids"]
            changes_summary["modify_rule_ids"] = rules_changes_summary["modify_rule_ids"]
            changes_summary["total_rule_ids"] = rules_changes_summary["total_rule_ids"]

            if args.output == display.OUTPUT_TYPE_JSON:
                output = get_upload_output_as_dict(
                    None,
                    test_results,
                    [],
                    args.verbose,
                    args.skip_tests,
                    UPLOAD_RESULT_SUCCESS,
                    changes_summary,
                )
                print(json.dumps(output, indent=display.JSON_INDENT_LEVEL))
            else:
                print_changes_summary(changes_summary)

        if not args.skip_summary and not args.dry_run:
            # if the user skips calculating the summary of the changes,
            # we don't show a message since there's no information shared with the user to act upon
            err = confirm("Would you like to make this change? [y/n]: ")
            if err is not None:
                return 0, ""

        if args.dry_run:
            return 0, ""

        # actually upload schemas
        upload_errored = manager.apply(args.verbose)
        if upload_errored:
            return 1, ""

        rule_upload_stats = run_rule_upload(
            backend=backend,
            session_id=session_id,
            verbose=args.verbose,
            output_type=args.output,
        )
    except BackendError as err:
        multi_err = BulkUploadDetectionsError.from_json(convert_unicode(err))
        if args.output == display.OUTPUT_TYPE_TEXT:
            print_upload_detection_error(multi_err)
        elif args.output == display.OUTPUT_TYPE_JSON:
            output = get_failed_upload_as_dict(
                multi_err,
                test_results,
                zip_info,
                args.verbose,
                args.skip_tests,
                UPLOAD_RESULT_FAILURE,
            )
            print(json.dumps(output, indent=display.JSON_INDENT_LEVEL))
        return 1, ""

    if args.output == display.OUTPUT_TYPE_JSON:
        output = get_upload_output_as_dict(
            rule_upload_stats,
            test_results,
            zip_info,
            args.verbose,
            args.skip_tests,
            UPLOAD_RESULT_SUCCESS,
            changes_summary,
        )
        print(json.dumps(output, indent=display.JSON_INDENT_LEVEL))
    else:
        print_upload_statistics(rule_upload_stats, changes_summary, pypanther_version)

    return 0, ""


def dry_run_rule_upload(backend: BackendClient, session_id: str, verbose: bool, output_type: str) -> ChangesSummary:
    if verbose and output_type == display.OUTPUT_TYPE_TEXT:
        print(cli_output.header("Calculating changes..."))
    elif output_type == display.OUTPUT_TYPE_TEXT:
        print("Calculating changes...")
        print()  # new line

    start_upload_response = backend.bulk_upload_detections(
        BulkUploadDetectionsParams(session_id=session_id, dry_run=True),
    )
    while True:
        time.sleep(1)
        status_response = backend.bulk_upload_detections_status(
            BulkUploadDetectionsStatusParams(job_id=start_upload_response.data.job_id),
        )
        if verbose and output_type == display.OUTPUT_TYPE_TEXT:
            print(f"Got status response {status_response.data.status}")
        if status_response.data.status == "Failed":
            raise BackendError(status_response.data.message)
        if status_response.data.status == "Succeeded":
            upload_stats = status_response.data.results
            if not upload_stats:
                raise BackendError("No results found in status response")
            changes_summary = ChangesSummary(
                new_rule_ids=upload_stats.new_rule_ids,
                delete_rule_ids=upload_stats.deleted_rule_ids,
                modify_rule_ids=upload_stats.modified_rule_ids,
                total_rule_ids=upload_stats.total_rule_ids,
                new_schema_names=[],
                modified_schema_names=[],
                existed_schema_names=[],
            )
            return changes_summary


def run_rule_upload(
    backend: BackendClient,
    session_id: str,
    verbose: bool,
    output_type: str,
) -> BulkUploadDetectionsResults:
    resp = backend.bulk_upload_detections(
        BulkUploadDetectionsParams(session_id=session_id, dry_run=False),
    )

    while True:
        time.sleep(1)
        status_response = backend.bulk_upload_detections_status(
            BulkUploadDetectionsStatusParams(job_id=resp.data.job_id),
        )
        if verbose and output_type == display.OUTPUT_TYPE_TEXT:
            print(f"Got status response {status_response.data.status}")
        if status_response.data.status == "Failed":
            raise BackendError(status_response.data.message)
        if status_response.data.status == "Succeeded":
            results = status_response.data.results
            if not results:
                raise BackendError("No results found in status response")
            return results


def zip_contents(named_temp_file: Any) -> list[zipfile.ZipInfo]:
    with zipfile.ZipFile(named_temp_file, "w") as zip_out:
        for root, dir_, files in os.walk("."):
            for bad in IGNORE_FOLDERS:
                if bad in dir_:
                    dir_.remove(bad)

            for file in files:
                if not fnmatch(file, "*.py"):
                    continue

                filepath = os.path.join(root, file)

                zip_out.write(
                    filepath,
                    arcname=filepath,
                )

        return zip_out.infolist()


# Uploads the zip file and returns the session_id
def upload_zip(
    backend: BackendClient,
    archive: str,
    verbose: bool,
    output_type: str,
    pypanther_version: str,
) -> str:
    if verbose and output_type == display.OUTPUT_TYPE_TEXT:
        print("requesting presigned URL for upload")
        print()

    response = backend.bulk_upload_presigned_url(
        params=BulkUploadPresignedURLParams(pypanther_version=pypanther_version),
    )

    with open(archive, "rb") as analysis_zip:
        if output_type == display.OUTPUT_TYPE_TEXT:
            if verbose:
                print(
                    f"Uploading detections zip file to URL: {response.data.detections_url} with pypanther version {pypanther_version}",
                )
            else:
                print("Uploading detections...")

        headers = {"x-amz-meta-pypantherversion": pypanther_version}
        data = base64.b64encode(analysis_zip.read()).decode("utf-8")
        # The timeout is set to 300 seconds to allow for larger files to be uploaded
        requests.put(response.data.detections_url, data=data, headers=headers, timeout=300)
        if verbose and output_type == display.OUTPUT_TYPE_TEXT:
            print("finished uploading detections zip file")

        return response.data.session_id


def confirm(warning_text: str) -> Optional[str]:
    warning_text = cli_output.warning(warning_text)
    choice = input(warning_text).lower()
    if choice != "y":
        print(cli_output.warning(f'Exiting upload due to entered response "{choice}"'))
        return "User did not confirm"

    print()  # new line
    return None


def get_upload_output_as_dict(
    upload_stats: BulkUploadDetectionsResults | None,
    test_results: testing.TestResults,
    zip_infos: list[zipfile.ZipInfo],
    verbose: bool,
    skip_tests: bool,
    upload_result: str,
    changes_summary: ChangesSummary | None,
) -> dict:
    output: dict[str, Any] = {"result": upload_result}
    if upload_stats:
        schema_stats = {}
        if changes_summary:
            schema_stats = {
                "new_schema_names": changes_summary["new_schema_names"],
                "modified_schema_names": changes_summary["modified_schema_names"],
                "existed_schema_names": changes_summary["existed_schema_names"],
            }
        output["upload_statistics"] = {
            "rules": asdict(upload_stats),
            "schemas": schema_stats,
        }
    if not skip_tests:
        output["tests"] = testing.test_output_dict(test_results, verbose)
    if verbose:
        output["registered_rules"] = [rule.id for rule in registered_rules()]
        if len(zip_infos) > 0:
            output["included_files"] = [info.filename for info in zip_infos]
    if changes_summary:
        output["summary"] = {
            "new_rule_ids": len(changes_summary["new_rule_ids"]),
            "delete_rule_ids": len(changes_summary["delete_rule_ids"]),
            "modify_rule_ids": len(changes_summary["modify_rule_ids"]),
            "new_schema_names": len(changes_summary["new_schema_names"]),
            "modified_schema_names": len(changes_summary["modified_schema_names"]),
            "existed_schema_names": len(changes_summary["existed_schema_names"]),
        }

    return output


def get_failed_upload_as_dict(
    err: BulkUploadDetectionsError,
    test_results: testing.TestResults,
    zip_infos: list[zipfile.ZipInfo],
    verbose: bool,
    skip_tests: bool,
    upload_result: str,
) -> dict:
    output = {"failed_upload_details": err.asdict(), "result": upload_result}
    if not skip_tests:
        output["tests"] = testing.test_output_dict(test_results, verbose)
    if verbose:
        output["registered_rules"] = [rule.id for rule in registered_rules()]
        if len(zip_infos) > 0:
            output["included_files"] = [info.filename for info in zip_infos]

    return output


def print_registered_rules() -> None:
    if len(registered_rules()) == 0:
        return

    print(cli_output.header("Registered Rules"))
    for i, rule in enumerate(registered_rules(), start=1):
        print(INDENT, f"{i}. {rule.id}")
    print()  # new line


def print_included_files(zip_info: list[zipfile.ZipInfo]) -> None:
    print(cli_output.header("Included files:"))
    for info in zip_info:
        print(INDENT, f"- {info.filename}")
    print()  # new line


def print_upload_statistics(
    rule_results: BulkUploadDetectionsResults,
    schema_results: ChangesSummary,
    pypanther_version: str,
) -> None:
    print(cli_output.header("Upload Statistics"))

    print(INDENT, f"{cli_output.bold('PyPanther Version:')} {pypanther_version}")
    print()  # new line

    print(INDENT, cli_output.bold("Rules:"))
    print(INDENT * 2, "{:<9} {:>4}".format("New:      ", len(rule_results.new_rule_ids)))
    print(INDENT * 2, "{:<9} {:>4}".format("Modified: ", len(rule_results.modified_rule_ids)))
    print(INDENT * 2, "{:<9} {:>4}".format("Deleted:  ", len(rule_results.deleted_rule_ids)))
    print(INDENT * 2, "{:<9} {:>4}".format("Total:    ", len(rule_results.total_rule_ids)))
    print()  # new line
    if (
        len(schema_results["new_schema_names"]) > 0
        or len(schema_results["modified_schema_names"]) > 0
        or len(schema_results["existed_schema_names"]) > 0
    ):
        print(INDENT, cli_output.bold("Schemas:"))
        print(INDENT * 2, "{:<9} {:>4}".format("New:          ", len(schema_results["new_schema_names"])))
        print(INDENT * 2, "{:<9} {:>4}".format("Modified:     ", len(schema_results["modified_schema_names"])))
        print(INDENT * 2, "{:<9} {:>4}".format("Not Modified: ", len(schema_results["existed_schema_names"])))
        print()  # new line


def print_upload_detection_error(err: BulkUploadDetectionsError) -> None:
    print(cli_output.failed("Upload Failed"))
    if err.error != "":
        print(INDENT, f"- {cli_output.failed(err.error)}")


def print_changes_summary(changes_summary: ChangesSummary) -> None:
    # Check if any of the keys starting with new/delete/modify have a value greater than 0
    if (
        changes_summary["new_rule_ids"]
        or changes_summary["delete_rule_ids"]
        or changes_summary["modify_rule_ids"]
        or changes_summary["new_schema_names"]
        or changes_summary["modified_schema_names"]
    ):
        print(cli_output.header("Changes"))
        print()  # new line
        if changes_summary["new_rule_ids"]:
            print(f"New Rules [{len(changes_summary['new_rule_ids'])}]:")
            for id_ in changes_summary["new_rule_ids"]:
                print(f"+ {id_}")
            print()  # new line
        if changes_summary["delete_rule_ids"]:
            print(f"Delete Rules [{len(changes_summary['delete_rule_ids'])}]:")
            for id_ in changes_summary["delete_rule_ids"]:
                print(f"- {id_}")
            print()  # new line
        if changes_summary["modify_rule_ids"]:
            print(f"Modify Rules [{len(changes_summary['modify_rule_ids'])}]:")
            for id_ in changes_summary["modify_rule_ids"]:
                print(f"~ {id_}")
            print()  # new line
        if changes_summary["new_schema_names"]:
            print(f"New Schemas [{len(changes_summary['new_schema_names'])}]:")
            for name in changes_summary["new_schema_names"]:
                print(f"+ {name}")
            print()  # new line
        if changes_summary["modified_schema_names"]:
            print(f"Modify Schemas [{len(changes_summary['modified_schema_names'])}]:")
            for name in changes_summary["modified_schema_names"]:
                print(f"~ {name}")

    print(cli_output.header("Changes Summary"))
    print(INDENT, f"New Rules:      {len(changes_summary['new_rule_ids']):>4}")
    print(INDENT, f"Modified Rules: {len(changes_summary['modify_rule_ids']):>4}")
    print(INDENT, f"Deleted Rules:  {len(changes_summary['delete_rule_ids']):>4}")
    print(INDENT, f"Total Rules:    {len(changes_summary['total_rule_ids']):>4}")
    print()  # new line
    if (
        len(changes_summary["new_schema_names"]) > 0
        or len(changes_summary["modified_schema_names"]) > 0
        or len(changes_summary["existed_schema_names"]) > 0
    ):
        print(INDENT, f"New Schemas:          {len(changes_summary['new_schema_names']):>4}")
        print(INDENT, f"Modified Schemas:     {len(changes_summary['modified_schema_names']):>4}")
        print(INDENT, f"Not Modified Schemas: {len(changes_summary['existed_schema_names']):>4}")
        print()  # new line



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional
from urllib.parse import urlparse

from .errors import is_retryable_error

if TYPE_CHECKING:
    # defer loading to improve performance
    from gql import Client as GraphQLClient
    from graphql import DocumentNode, ExecutionResult

from gql.transport.exceptions import TransportQueryError

from pypanther import display

from .client import (
    BackendCheckResponse,
    BackendError,
    BackendResponse,
    BulkUploadDetectionsParams,
    BulkUploadDetectionsResponse,
    BulkUploadDetectionsResults,
    BulkUploadDetectionsStatusParams,
    BulkUploadDetectionsStatusResponse,
    BulkUploadPresignedURLParams,
    BulkUploadPresignedURLResponse,
    Client,
    ListSchemasParams,
    ListSchemasResponse,
    PermanentBackendError,
    Schema,
    UpdateSchemaParams,
    UpdateSchemaResponse,
)


@dataclass(frozen=True)
class PublicAPIClientOptions:
    host: str
    token: str
    user_id: str
    verbose: bool
    output_type: str


class PublicAPIRequests:  # pylint: disable=too-many-public-methods
    _cache: Dict[str, str]

    def __init__(self) -> None:
        self._cache = {}

    def version_query(self) -> "DocumentNode":
        return self._load("get_version")

    def bulk_upload_presigned_url_query(self) -> "DocumentNode":
        return self._load("bulk_upload_presigned_url")

    def bulk_upload_detections_mutation(self) -> "DocumentNode":
        return self._load("bulk_upload_detections")

    def bulk_upload_detections_status_query(self) -> "DocumentNode":
        return self._load("bulk_upload_detections_status")

    def list_schemas_query(self) -> "DocumentNode":
        return self._load("list_schemas")

    def update_schema_mutation(self) -> "DocumentNode":
        return self._load("create_or_update_schema")

    def _load(self, name: str) -> "DocumentNode":
        # defer loading to improve performance
        from gql import gql

        if name not in self._cache:
            self._cache[name] = Path(_get_graphql_content_filepath(name)).read_text(encoding="utf-8")

        return gql(self._cache[name])


class PublicAPIClient(Client):  # pylint: disable=too-many-public-methods
    _user_id: str
    _requests: PublicAPIRequests
    _gql_client: "GraphQLClient"

    def __init__(self, opts: PublicAPIClientOptions):
        self._user_id = opts.user_id
        self._requests = PublicAPIRequests()
        self._gql_client = _build_client(opts.host, opts.token, opts.verbose, opts.output_type)

    def check(self) -> BackendCheckResponse:
        res = self._execute(self._requests.version_query())

        if res.errors:
            for err in res.errors:
                logging.error(err.message)

            return BackendCheckResponse(success=False, message="connection check failed")

        if res.data is None:
            return BackendCheckResponse(success=False, message="backend sent empty response")

        panther_version = res.data.get("generalSettings", {}).get("pantherVersion")
        if panther_version is None:
            return BackendCheckResponse(
                success=False,
                message="did not receive version in response",
            )

        return BackendCheckResponse(success=True, message=f"connected to Panther backend on version: {panther_version}")

    def bulk_upload_presigned_url(
        self,
        params: BulkUploadPresignedURLParams,
    ) -> BackendResponse[BulkUploadPresignedURLResponse]:
        query = self._requests.bulk_upload_presigned_url_query()
        request_params = {
            "input": {
                "pypantherVersion": params.pypanther_version,
            },
        }
        response = self._safe_execute(query, variable_values=request_params)
        url = response.data.get("bulkUploadPresignedUrl", {}).get("detectionsURL")  # type: ignore
        session_id = response.data.get("bulkUploadPresignedUrl", {}).get("sessionId")  # type: ignore
        return BackendResponse(
            status_code=200,
            data=BulkUploadPresignedURLResponse(
                detections_url=url,
                session_id=session_id,
            ),
        )

    def bulk_upload_detections(
        self,
        params: BulkUploadDetectionsParams,
    ) -> BackendResponse[BulkUploadDetectionsResponse]:
        query = self._requests.bulk_upload_detections_mutation()
        upload_params = {
            "input": {
                "dryRun": params.dry_run,
                "sessionId": params.session_id,
            },
        }
        res = self._safe_execute(query, variable_values=upload_params)
        job_id = res.data.get("bulkUploadDetections", {}).get("id")  # type: ignore
        return BackendResponse(
            status_code=200,
            data=BulkUploadDetectionsResponse(job_id=job_id),
        )

    def bulk_upload_detections_status(
        self,
        params: BulkUploadDetectionsStatusParams,
    ) -> BackendResponse[BulkUploadDetectionsStatusResponse]:
        query = self._requests.bulk_upload_detections_status_query()
        upload_params = {"input": params.job_id}
        res = self._safe_execute(query, variable_values=upload_params).data.get("bulkUploadDetectionsStatus", {})  # type: ignore
        results = None
        if res_results := res.get("results"):
            results = BulkUploadDetectionsResults(
                new_rule_ids=res_results.get("newRuleIds") or [],
                modified_rule_ids=res_results.get("modifiedRuleIds") or [],
                deleted_rule_ids=res_results.get("deletedRuleIds") or [],
                total_rule_ids=res_results.get("totalRuleIds") or [],
            )

        return BackendResponse(
            status_code=200,
            data=BulkUploadDetectionsStatusResponse(
                status=res.get("status"),
                message=res.get("message"),
                results=results,
            ),
        )

    def list_schemas(self, params: ListSchemasParams) -> BackendResponse[ListSchemasResponse]:
        gql_params = {
            "input": {
                "isManaged": params.is_managed,
            },
        }
        res = self._execute(self._requests.list_schemas_query(), gql_params)
        if res.errors:
            for err in res.errors:
                logging.error(err.message)
            raise BackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        schemas = []
        for edge in res.data.get("schemas", {}).get("edges", []):
            node = edge.get("node", {})
            schema = Schema(
                created_at=node.get("createdAt", ""),
                description=node.get("description", ""),
                is_managed=node.get("isManaged", False),
                name=node.get("name", ""),
                reference_url=node.get("referenceURL", ""),
                revision=node.get("revision", ""),
                spec=node.get("spec", ""),
                updated_at=node.get("updatedAt", ""),
                field_discovery_enabled=node.get("fieldDiscoveryEnabled", False),
            )
            schemas.append(schema)

        return BackendResponse(status_code=200, data=ListSchemasResponse(schemas=schemas))

    def update_schema(self, params: UpdateSchemaParams) -> BackendResponse:
        gql_params = {
            "input": {
                "description": params.description,
                "name": params.name,
                "referenceURL": params.reference_url,
                "revision": params.revision,
                "spec": params.spec,
                "isFieldDiscoveryEnabled": params.field_discovery_enabled,
            },
        }
        try:
            res = self._execute(self._requests.update_schema_mutation(), gql_params)
            if res.errors:
                for err in res.errors:
                    logging.error(err.message)
                raise BackendError(res.errors)
        except TransportQueryError as exc:
            raise BackendError(exc)

        if res.data is None:
            raise BackendError("empty data")

        schema = res.data.get("schema", {})
        return BackendResponse(
            status_code=200,
            data=UpdateSchemaResponse(
                schema=Schema(
                    created_at=schema.get("createdAt", ""),
                    description=schema.get("description", ""),
                    is_managed=schema.get("isManaged", False),
                    name=schema.get("name", ""),
                    reference_url=schema.get("referenceURL", ""),
                    revision=schema.get("revision", ""),
                    spec=schema.get("spec", ""),
                    updated_at=schema.get("updatedAt", ""),
                    field_discovery_enabled=schema.get("fieldDiscoveryEnabled", False),
                ),
            ),
        )

    def _execute(
        self,
        request: "DocumentNode",
        variable_values: Optional[Dict[str, Any]] = None,
    ) -> "ExecutionResult":
        return self._gql_client.execute(request, variable_values=variable_values, get_execution_result=True)

    def _safe_execute(
        self,
        request: "DocumentNode",
        variable_values: Optional[Dict[str, Any]] = None,
    ) -> "ExecutionResult":
        try:
            res = self._execute(request, variable_values=variable_values)
        except TransportQueryError as e:  # pylint: disable=C0103
            err = PermanentBackendError(e)
            if e.errors and len(e.errors) > 0:
                err = BackendError(e.errors[0])  # type: ignore
                err.permanent = not is_retryable_error(e.errors[0])
            raise err from e

        if res.errors:
            raise PermanentBackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        return res


_API_URL_PATH = "public/graphql"
_API_DOMAIN_PREFIX = "api"
_API_TOKEN_HEADER = "X-API-Key"  # nosec


def _build_client(host: str, token: str, verbose: bool, output_type: str = display.OUTPUT_TYPE_TEXT) -> "GraphQLClient":
    from gql import Client as GraphQLClient
    from gql.transport.aiohttp import AIOHTTPTransport

    graphql_url = _build_api_url(host)
    if verbose and output_type == display.OUTPUT_TYPE_TEXT:
        print("Panther Public API endpoint: %s", graphql_url)
        print()  # new line

    transport = AIOHTTPTransport(url=graphql_url, headers={_API_TOKEN_HEADER: token})

    return GraphQLClient(transport=transport, fetch_schema_from_transport=False, execute_timeout=30)


def is_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def _build_api_url(host: str) -> str:
    if is_url(host):
        return host

    return f"https://{_API_DOMAIN_PREFIX}.{host}/{_API_URL_PATH}"


def _get_graphql_content_filepath(name: str) -> str:
    work_dir = os.path.dirname(__file__)
    return os.path.join(work_dir, "graphql", f"{name}.graphql")



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.wiz import wiz_actor, wiz_alert_context, wiz_success


@panther_managed
class WizUpdateScannerSettings(Rule):
    id = "Wiz.Update.Scanner.Settings-prototype"
    default_description = "This rule detects updates of Wiz scanner settings."
    display_name = "Wiz Update Scanner Settings"
    default_runbook = (
        "Verify that this change was planned. If not, revert the change and ensure this doesn't happen again."
    )
    default_reference = "https://www.wiz.io/academy/secret-scanning"
    default_severity = Severity.MEDIUM
    reports = {"MITRE ATT&CK": ["TA0005:T1562.001"]}
    log_types = [LogType.WIZ_AUDIT]

    def rule(self, event):
        if not wiz_success(event):
            return False
        return event.get("action", "ACTION_NOT_FOUND") == "UpdateScannerSettings"

    def title(self, event):
        actor = wiz_actor(event)
        return f"[Wiz]: [{event.get('action', 'ACTION_NOT_FOUND')}] action performed by {actor.get('type')} [{actor.get('name')}]"

    def dedup(self, event):
        return event.get("id")

    def alert_context(self, event):
        return wiz_alert_context(event)

    tests = [
        RuleTest(
            name="UpdateScannerSettings",
            expected_result=True,
            log={
                "id": "dd48b7fe-576d-453d-a0d0-1f61425b1bb7",
                "action": "UpdateScannerSettings",
                "requestId": "d5c55350-0d54-46eb-88ee-4942f80e700c",
                "status": "SUCCESS",
                "timestamp": "2024-06-18T12:09:33.985762Z",
                "actionParameters": {
                    "input": {
                        "patch": {
                            "computeResourceGroupMemberScanSamplingEnabled": True,
                            "maxComputeResourceGroupMemberScanCount": 2,
                            "prioritizeActiveComputeResourceGroupMembers": True,
                        },
                    },
                    "selection": [
                        "__typename",
                        {
                            "scannerSettings": [
                                "__typename",
                                "computeResourceGroupMemberScanSamplingEnabled",
                                "maxComputeResourceGroupMemberScanCount",
                                {"customFileDetectionList": ["__typename", "id", "url", "fileDetectionCount"]},
                            ],
                        },
                    ],
                },
                "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
                "sourceIP": "12.34.56.78",
                "serviceAccount": None,
                "user": {"id": "test.user@company.com", "name": "user@company.com"},
            },
        ),
        RuleTest(
            name="CreateUser",
            expected_result=False,
            log={
                "id": "220d23be-f07c-4d97-b4a6-87ad04eddb14",
                "action": "CreateUser",
                "requestId": "0d9521b2-c3f8-4a73-bf7c-20257788752e",
                "status": "SUCCESS",
                "timestamp": "2024-07-29T09:40:15.66643Z",
                "actionParameters": {
                    "input": {
                        "assignedProjectIds": None,
                        "email": "testy@company.com",
                        "expiresAt": None,
                        "name": "Test User",
                        "role": "GLOBAL_ADMIN",
                    },
                    "selection": ["__typename", {"user": ["__typename", "id"]}],
                },
                "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36",
                "sourceIP": "8.8.8.8",
                "serviceAccount": None,
                "user": {"id": "someuser@company.com", "name": "someuser@company.com"},
            },
        ),
        RuleTest(
            name="UpdateScannerSettings - Fail",
            expected_result=False,
            log={
                "id": "dd48b7fe-576d-453d-a0d0-1f61425b1bb7",
                "action": "UpdateScannerSettings",
                "requestId": "d5c55350-0d54-46eb-88ee-4942f80e700c",
                "status": "FAILED",
                "timestamp": "2024-06-18T12:09:33.985762Z",
                "actionParameters": {},
                "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
                "sourceIP": "12.34.56.78",
                "serviceAccount": None,
                "user": {"id": "test.user@company.com", "name": "user@company.com"},
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.gcp import gcp_alert_context, get_binding_deltas


@panther_managed
class GCPComputeIAMPolicyUpdate(Rule):
    display_name = "GCP Compute IAM Policy Update Detection"
    id = "GCP.Compute.IAM.Policy.Update-prototype"
    default_severity = Severity.MEDIUM
    log_types = [LogType.GCP_AUDIT_LOG]
    default_description = "This rule detects updates to IAM policies for Compute Disks, Images, and Snapshots.\n"
    default_runbook = (
        "Ensure that the IAM policy update was expected. Unauthorized changes can lead to security risks.\n"
    )
    default_reference = "https://cloud.google.com/compute/docs/access/iam"
    SUSPICIOUS_ACTIONS = [
        "v1.compute.disks.setIamPolicy",
        "v1.compute.images.setIamPolicy",
        "v1.compute.snapshots.setIamPolicy",
    ]

    def rule(self, event):
        if event.deep_get("protoPayload", "response", "error"):
            return False
        method = event.deep_get("protoPayload", "methodName", default="METHOD_NOT_FOUND")
        if method in self.SUSPICIOUS_ACTIONS:
            return True
        return False

    def title(self, event):
        actor = event.deep_get("protoPayload", "authenticationInfo", "principalEmail", default="<ACTOR_NOT_FOUND>")
        items = event.deep_get("protoPayload", "methodName", default="ITEMS_NOT_FOUND. ").split(".")[-2]
        project = event.deep_get("resource", "labels", "project_id", default="<PROJECT_NOT_FOUND>")
        return f"[GCP]: [{actor}] updated IAM policy for [{items}] on project [{project}]"

    def alert_context(self, event):
        context = gcp_alert_context(event)
        service_accounts = event.deep_get("protoPayload", "request", "serviceAccounts")
        if not service_accounts:
            service_account_emails = "<SERVICE_ACCOUNT_EMAILS_NOT_FOUND>"
        else:
            service_account_emails = [service_acc["email"] for service_acc in service_accounts]
        context["serviceAccount"] = service_account_emails
        context["binding_deltas"] = get_binding_deltas(event)
        return context

    tests = [
        RuleTest(
            name="IAM policy update on a Compute Disk",
            expected_result=True,
            log={
                "insertId": "1abcd23efg456",
                "labels": {"compute.googleapis.com/root_trigger_id": "trigger-id-1"},
                "logName": "projects/test-project/logs/cloudaudit.googleapis.com%2Factivity",
                "protoPayload": {
                    "@type": "type.googleapis.com/google.cloud.audit.AuditLog",
                    "authenticationInfo": {
                        "principalEmail": "user@example.com",
                        "principalSubject": "serviceAccount:user@example.com",
                        "serviceAccountKeyName": "//iam.googleapis.com/projects/test-project/serviceAccounts/user@example.com/keys/key-id",
                    },
                    "authorizationInfo": [
                        {
                            "granted": True,
                            "permission": "compute.disks.setIamPolicy",
                            "resource": "projects/test-project/zones/us-central1-a/disks/disk-1",
                            "resourceAttributes": {
                                "name": "projects/test-project/zones/us-central1-a/disks/disk-1",
                                "service": "compute",
                                "type": "compute.disks",
                            },
                        },
                    ],
                    "methodName": "v1.compute.disks.setIamPolicy",
                    "request": {
                        "@type": "type.googleapis.com/compute.disks.setIamPolicy",
                        "policy": {"bindings": [{"members": ["user:anonymized@example.com"], "role": "roles/owner"}]},
                    },
                    "requestMetadata": {
                        "callerIP": "192.0.2.1",
                        "callerSuppliedUserAgent": "google-cloud-sdk",
                        "destinationAttributes": {},
                        "requestAttributes": {"auth": {}, "time": "2023-10-01T12:34:56.789Z"},
                    },
                    "resourceLocation": {"currentLocations": ["us-central1-a"]},
                    "resourceName": "projects/test-project/zones/us-central1-a/disks/disk-1",
                    "serviceName": "compute.googleapis.com",
                },
                "receiveTimestamp": "2023-10-01T12:34:57.123Z",
                "resource": {
                    "labels": {"disk_id": "disk-id-1", "project_id": "test-project", "zone": "us-central1-a"},
                    "type": "gce_disk",
                },
                "severity": "NOTICE",
                "timestamp": "2023-10-01T12:34:56.789Z",
            },
        ),
        RuleTest(
            name="IAM policy update on a Compute Image",
            expected_result=True,
            log={
                "insertId": "2hijk34lmn789",
                "labels": {"compute.googleapis.com/root_trigger_id": "trigger-id-2"},
                "logName": "projects/test-project/logs/cloudaudit.googleapis.com%2Factivity",
                "protoPayload": {
                    "@type": "type.googleapis.com/google.cloud.audit.AuditLog",
                    "authenticationInfo": {
                        "principalEmail": "user@example.com",
                        "principalSubject": "serviceAccount:user@example.com",
                        "serviceAccountKeyName": "//iam.googleapis.com/projects/test-project/serviceAccounts/user@example.com/keys/key-id",
                    },
                    "authorizationInfo": [
                        {
                            "granted": True,
                            "permission": "compute.images.setIamPolicy",
                            "resource": "projects/test-project/global/images/image-1",
                            "resourceAttributes": {
                                "name": "projects/test-project/global/images/image-1",
                                "service": "compute",
                                "type": "compute.images",
                            },
                        },
                    ],
                    "methodName": "v1.compute.images.setIamPolicy",
                    "request": {
                        "@type": "type.googleapis.com/compute.images.setIamPolicy",
                        "policy": {"bindings": [{"members": ["user:anonymized@example.com"], "role": "roles/owner"}]},
                    },
                    "requestMetadata": {
                        "callerIP": "192.0.2.2",
                        "callerSuppliedUserAgent": "google-cloud-sdk",
                        "destinationAttributes": {},
                        "requestAttributes": {"auth": {}, "time": "2023-10-01T12:45:56.789Z"},
                    },
                    "resourceLocation": {"currentLocations": ["us-central1-a"]},
                    "resourceName": "projects/test-project/global/images/image-1",
                    "serviceName": "compute.googleapis.com",
                },
                "receiveTimestamp": "2023-10-01T12:45:57.123Z",
                "resource": {"labels": {"image_id": "image-id-1", "project_id": "test-project"}, "type": "gce_image"},
                "severity": "NOTICE",
                "timestamp": "2023-10-01T12:45:56.789Z",
            },
        ),
        RuleTest(
            name="Non-IAM policy update on a Compute Disk",
            expected_result=False,
            log={
                "insertId": "4stuv78wxy345",
                "labels": {"compute.googleapis.com/root_trigger_id": "trigger-id-4"},
                "logName": "projects/test-project/logs/cloudaudit.googleapis.com%2Factivity",
                "protoPayload": {
                    "@type": "type.googleapis.com/google.cloud.audit.AuditLog",
                    "authenticationInfo": {
                        "principalEmail": "user@example.com",
                        "principalSubject": "serviceAccount:user@example.com",
                        "serviceAccountKeyName": "//iam.googleapis.com/projects/test-project/serviceAccounts/user@example.com/keys/key-id",
                    },
                    "authorizationInfo": [
                        {
                            "granted": True,
                            "permission": "compute.disks.create",
                            "resource": "projects/test-project/zones/us-central1-a/disks/disk-2",
                            "resourceAttributes": {
                                "name": "projects/test-project/zones/us-central1-a/disks/disk-2",
                                "service": "compute",
                                "type": "compute.disks",
                            },
                        },
                    ],
                    "methodName": "v1.compute.disks.create",
                    "requestMetadata": {
                        "callerIP": "192.0.2.4",
                        "callerSuppliedUserAgent": "google-cloud-sdk",
                        "destinationAttributes": {},
                        "requestAttributes": {"auth": {}, "time": "2023-10-01T13:00:00.000Z"},
                    },
                    "resourceLocation": {"currentLocations": ["us-central1-a"]},
                    "resourceName": "projects/test-project/zones/us-central1-a/disks/disk-2",
                    "serviceName": "compute.googleapis.com",
                },
                "receiveTimestamp": "2023-10-01T13:00:01.000Z",
                "resource": {
                    "labels": {"disk_id": "disk-id-2", "project_id": "test-project", "zone": "us-central1-a"},
                    "type": "gce_disk",
                },
                "severity": "NOTICE",
                "timestamp": "2023-10-01T13:00:00.000Z",
            },
        ),
    ]



from pypanther.rules.dropbox.dropbox_admin_sign_in_as_session import (
    DropboxAdminsigninasSession as DropboxAdminsigninasSession,
)
from pypanther.rules.dropbox.dropbox_external_share import DropboxExternalShare as DropboxExternalShare
from pypanther.rules.dropbox.dropbox_linked_team_application_added import (
    DropboxLinkedTeamApplicationAdded as DropboxLinkedTeamApplicationAdded,
)
from pypanther.rules.dropbox.dropbox_ownership_transfer import DropboxOwnershipTransfer as DropboxOwnershipTransfer
from pypanther.rules.dropbox.dropbox_user_disabled_2fa import DropboxUserDisabled2FA as DropboxUserDisabled2FA



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.aws import aws_guardduty_context


@panther_managed
class AWSGuardDutyCriticalSeverityFinding(Rule):
    id = "AWS.GuardDuty.CriticalSeverityFinding-prototype"
    display_name = "AWS GuardDuty Critical Severity Finding"
    log_types = [LogType.AWS_GUARDDUTY]
    tags = ["AWS"]
    default_severity = Severity.CRITICAL
    dedup_period_minutes = 15
    default_description = "A critical-severity GuardDuty finding has been identified.\n"
    default_runbook = "Search related logs to understand the root cause of the activity. Search the Panther Summary Attribute type value in https://docs.aws.amazon.com/guardduty/latest/ug/guardduty_finding-types-active.html for additional details.\n"
    default_reference = (
        "https://docs.aws.amazon.com/guardduty/latest/ug/guardduty_findings.html#guardduty_findings-severity"
    )
    summary_attributes = ["severity", "type", "title", "p_any_domain_names", "p_any_aws_arns", "p_any_aws_account_ids"]

    def rule(self, event):
        if event.deep_get("service", "additionalInfo", "sample"):
            # in case of sample data
            # https://docs.aws.amazon.com/guardduty/latest/ug/sample_findings.html
            return False
        return 9.0 <= float(event.get("severity", 0)) <= 10.0

    def title(self, event):
        return event.get("title")

    def alert_context(self, event):
        return aws_guardduty_context(event)

    tests = [
        RuleTest(
            name="Critical Sev Finding",
            expected_result=True,
            log={
                "accountId": "123456789012",
                "arn": "arn:aws:guardduty:us-east-1:123456789012:detector/111111bbbbbbbbbb5555555551111111/finding/8c258d654378414b9bc26f61a93eb816",
                "createdAt": "2025-02-25 20:40:06.721",
                "description": "A sequence of actions involving 4 signals indicating a possible credential compromise was observed for IAMUser/john_doe with principalId AIDA3UBBJ2K3TVEXAMPLE in account 111122223333\nbetween eventFirstSeen and eventLastSeen with the following behaviors:\n  - 5 MITRE ATT&CK tactics observed: Persistence, Privilege Escalation, Defense Evasion, Discovery, Initial Access\n  - 5 MITRE ATT&CK techniques observed: T1562.008 - Impair Defenses: Disable or Modify Cloud Logs, T1098.003 - Account Manipulation: Additional Cloud Roles, T1078.004 - Valid Accounts: Cloud Accounts, T1087.004 - Account Discovery: Cloud Account, T1098 - Account Manipulation\n  - Connected from a known Tor Exit Node: 10.0.0.1\n  - 4 sensitive APIs called: cloudtrail:DeleteTrail, iam:AttachRolePolicy, iam:CreateRole, iam:ListUsers\n",
                "id": "8c258d654378414b9bc26f61a93eb816",
                "partition": "aws",
                "region": "us-east-1",
                "resource": {"resourceType": "AttackSequence"},
                "schemaVersion": "2.0",
                "service": {
                    "additionalInfo": {},
                    "archived": False,
                    "count": 1,
                    "detectorId": "111111bbbbbbbbbb5555555551111111",
                    "eventFirstSeen": "2025-02-25 20:40:06.000000000",
                    "eventLastSeen": "2025-02-25 20:40:06.000000000",
                    "featureName": "Correlation",
                    "resourceRole": "TARGET",
                    "serviceName": "guardduty",
                },
                "severity": 9,
                "title": "Potential credential compromise of IAMUser/john_doe indicated by a sequence of actions.",
                "type": "AttackSequence:IAM/CompromisedCredentials",
                "updatedAt": "2025-02-25 20:40:06.721",
            },
        ),
        RuleTest(
            name="Critical Sev Finding As Sample Data",
            expected_result=False,
            log={
                "accountId": "123456789012",
                "arn": "arn:aws:guardduty:us-east-1:123456789012:detector/111111bbbbbbbbbb5555555551111111/finding/b174b38d16e34da0a4c66012083bd8f7",
                "createdAt": "2025-02-25 20:40:06.765",
                "description": "A sequence of actions involving 14 signals indicating a possible credential compromise one or more S3 bucket(s) was observed for IAMUser/john_doe with principalId AIDA3UBBJ2K3TVEXAMPLE in account 111122223333\nbetween eventFirstSeen and eventLastSeen with the following behaviors:\n  - 5 MITRE ATT&CK tactics observed: Exfiltration, Impact, Persistence, Defense Evasion, Discovery\n  - 5 MITRE ATT&CK techniques observed: T1526 - Cloud Service Discovery, T1098 - Account Manipulation, T1078.004 - Valid Accounts: Cloud Accounts, T1485 - Data Destruction, T1530 - Data from Cloud Storage\n  - Connected from a known Tor Exit Node: 10.0.0.1\n  - 7 sensitive APIs called: s3:DeleteObject, s3:GetObject, s3:PutBucketPublicAccessBlock, cloudtrail:DeleteTrail, iam:AttachUserPolicy, s3:ListObjects, s3:ListBuckets\n",
                "id": "b174b38d16e34da0a4c66012083bd8f7",
                "partition": "aws",
                "region": "us-east-1",
                "resource": {"resourceType": "AttackSequence"},
                "schemaVersion": "2.0",
                "service": {
                    "additionalInfo": {"sample": True, "type": "default", "value": '{"sample":true}'},
                    "archived": False,
                    "count": 1,
                    "detectorId": "111111bbbbbbbbbb5555555551111111",
                    "eventFirstSeen": "2025-02-25 20:40:06.000000000",
                    "eventLastSeen": "2025-02-25 20:40:06.000000000",
                    "featureName": "Correlation",
                    "resourceRole": "TARGET",
                    "serviceName": "guardduty",
                },
                "severity": 9,
                "title": "Potential data compromise of one or more S3 buckets involving a sequence of actions associated with IAMUser/john_doe.",
                "type": "AttackSequence:S3/CompromisedData",
                "updatedAt": "2025-02-25 20:40:06.765",
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.crowdstrike_event_streams import audit_keys_dict, cs_alert_context


@panther_managed
class CrowdstrikeAdminRoleAssigned(Rule):
    id = "Crowdstrike.AdminRoleAssigned-prototype"
    display_name = "Crowdstrike Admin Role Assigned"
    log_types = [LogType.CROWDSTRIKE_EVENT_STREAMS]
    default_severity = Severity.MEDIUM
    reports = {"MITRE ATT&CK": ["TA0003:T1098.003", "TA0004:T1098.003"]}
    default_description = "A user was assigned a priviledged role"
    default_runbook = "Confirm the role assignment is justified."
    # List of priviledged roles.
    # IMPORTANT: YOU MUST ADD ANY CUSTOM ADMIN ROLES YOURSELF
    # NG SIEM Admin
    # Remote Responder Admin
    ADMIN_ROLES = {
        "billing_dashboard_admin",
        "falconhost_admin",
        "firewall_manager",
        "xdr_admin",
        "remote_responder_three",
    }

    def get_roles_assigned(self, event):
        """Returns a list of the roles assigned in this event."""
        # Extract the AuditKeyValues construct
        audit_keys = audit_keys_dict(event)
        # Return Roles
        return audit_keys.get("roles", "").split(",")

    def rule(self, event):
        # Ignore non role-granting events
        if not all([event.deep_get("event", "OperationName") == "grantUserRoles", event.deep_get("event", "Success")]):
            return False
        # Raise alert if any of the admin roles were assigned
        roles_assigned = self.get_roles_assigned(event)
        return bool(self.ADMIN_ROLES & set(roles_assigned))

    def title(self, event):
        audit_keys = audit_keys_dict(event)
        actor = audit_keys["actor_user"]
        target = audit_keys["target_name"]
        admin_roles = set(self.get_roles_assigned(event)) & self.ADMIN_ROLES
        return f"{actor} assigned admin roles to {target}: {', '.join(list(admin_roles))}"

    def dedup(self, event):
        # The title includes the role names, but if the actor assigned more roles to the user, we
        #   dedup those alerts as well.
        audit_keys = audit_keys_dict(event)
        actor = audit_keys["actor_user"]
        target = audit_keys["target_name"]
        return f"{actor}-{target}"

    def alert_context(self, event):
        context = cs_alert_context(event)
        actor = context.get("actor_user", "UNKNWON_ACTOR")
        target = context.get("target_name", "UNKNOWN_TARGET")
        context["actor_target"] = f"{actor}-{target}"
        return context

    tests = [
        RuleTest(
            name="Admin Role Assigned (Single)",
            expected_result=True,
            log={
                "event": {
                    "AuditKeyValues": [
                        {"Key": "target_name", "ValueString": "merry.brandybuck@hobbiton.co"},
                        {"Key": "target_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                        {"Key": "target_cid", "ValueString": "fake_customer_id"},
                        {"Key": "roles", "ValueString": "billing_dashboard_admin"},
                        {"Key": "actor_cid", "ValueString": "fake_customer_id"},
                        {"Key": "trace_id", "ValueString": "897d300ad09137b362ee6a62846a9277"},
                        {"Key": "actor_user", "ValueString": "peregrin.took@hobbiton.co"},
                        {"Key": "actor_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                    ],
                    "OperationName": "grantUserRoles",
                    "ServiceName": "Crowdstrike Authentication",
                    "Success": True,
                    "UTCTimestamp": "2024-07-22 21:32:49.000000000",
                    "UserId": "peregrin.took@hobbiton.co",
                    "UserIp": "1.1.1.1",
                },
                "metadata": {
                    "customerIDString": "fake_customer_id",
                    "eventCreationTime": "2024-07-22 21:32:49.531000000",
                    "eventType": "AuthActivityAuditEvent",
                    "offset": 342905,
                    "version": "1.0",
                },
            },
        ),
        RuleTest(
            name="Admin Role Assigned (Multiple)",
            expected_result=True,
            log={
                "event": {
                    "AuditKeyValues": [
                        {"Key": "target_name", "ValueString": "merry.brandybuck@hobbiton.co"},
                        {"Key": "target_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                        {"Key": "target_cid", "ValueString": "fake_customer_id"},
                        {
                            "Key": "roles",
                            "ValueString": "custom_non_admin_role,billing_dashboard_admin,falconhost_admin",
                        },
                        {"Key": "actor_cid", "ValueString": "fake_customer_id"},
                        {"Key": "trace_id", "ValueString": "897d300ad09137b362ee6a62846a9277"},
                        {"Key": "actor_user", "ValueString": "peregrin.took@hobbiton.co"},
                        {"Key": "actor_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                    ],
                    "OperationName": "grantUserRoles",
                    "ServiceName": "Crowdstrike Authentication",
                    "Success": True,
                    "UTCTimestamp": "2024-07-22 21:32:49.000000000",
                    "UserId": "peregrin.took@hobbiton.co",
                    "UserIp": "1.1.1.1",
                },
                "metadata": {
                    "customerIDString": "fake_customer_id",
                    "eventCreationTime": "2024-07-22 21:32:49.531000000",
                    "eventType": "AuthActivityAuditEvent",
                    "offset": 342905,
                    "version": "1.0",
                },
            },
        ),
        RuleTest(
            name="Non-Admin Role Assigned",
            expected_result=False,
            log={
                "event": {
                    "AuditKeyValues": [
                        {"Key": "target_name", "ValueString": "merry.brandybuck@hobbiton.co"},
                        {"Key": "target_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                        {"Key": "target_cid", "ValueString": "fake_customer_id"},
                        {"Key": "roles", "ValueString": "custom_non_admin_role"},
                        {"Key": "actor_cid", "ValueString": "fake_customer_id"},
                        {"Key": "trace_id", "ValueString": "897d300ad09137b362ee6a62846a9277"},
                        {"Key": "actor_user", "ValueString": "peregrin.took@hobbiton.co"},
                        {"Key": "actor_user_uuid", "ValueString": "e70e5306-4a83-4a9f-9b59-a78c304c438b"},
                    ],
                    "OperationName": "grantUserRoles",
                    "ServiceName": "Crowdstrike Authentication",
                    "Success": True,
                    "UTCTimestamp": "2024-07-22 21:32:49.000000000",
                    "UserId": "peregrin.took@hobbiton.co",
                    "UserIp": "1.1.1.1",
                },
                "metadata": {
                    "customerIDString": "fake_customer_id",
                    "eventCreationTime": "2024-07-22 21:32:49.531000000",
                    "eventType": "AuthActivityAuditEvent",
                    "offset": 342905,
                    "version": "1.0",
                },
            },
        ),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed


@panther_managed
class AsanaWorkspaceGuestInvitePermissionsAnyone(Rule):
    default_description = "Typically inviting guests to Asana is permitted by few users. Enabling anyone to invite guests can potentially lead to unauthorized users gaining access to Asana."
    display_name = "Asana Workspace Guest Invite Permissions Anyone"
    default_reference = "https://help.asana.com/hc/en-us/articles/14109494654875-Admin-console#:~:text=Google%20SSO%20password.-,Guest%20invite%20controls,-Super%20admins%20of"
    default_severity = Severity.LOW
    log_types = [LogType.ASANA_AUDIT]
    id = "Asana.Workspace.Guest.Invite.Permissions.Anyone-prototype"

    def rule(self, event):
        return (
            event.get("event_type") == "workspace_guest_invite_permissions_changed"
            and event.deep_get("details", "new_value") == "anyone"
        )

    def title(self, event):
        workspace = event.deep_get("resource", "name", default="<WORKSPACE_NOT_FOUND>")
        actor = event.deep_get("actor", "email", default="<ACTOR_NOT_FOUND>")
        return f"Asana Workspace [{workspace}] guest invite permissions changed to anyone by [{actor}]."

    tests = [
        RuleTest(
            name="Anyone Allowed Guest Invite",
            expected_result=True,
            log={
                "actor": {
                    "actor_type": "user",
                    "email": "homer.simpson@example.io",
                    "gid": "12345",
                    "name": "Homer Simpson",
                },
                "context": {
                    "client_ip_address": "12.12.12.12",
                    "context_type": "web",
                    "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
                },
                "created_at": "2022-12-16 19:30:26.15",
                "details": {"new_value": "anyone", "old_value": "admins_only"},
                "event_category": "admin_settings",
                "event_type": "workspace_guest_invite_permissions_changed",
                "gid": "12345",
                "resource": {"gid": "12345", "name": "Example IO", "resource_type": "workspace"},
            },
        ),
        RuleTest(
            name="Other",
            expected_result=False,
            log={
                "actor": {
                    "actor_type": "user",
                    "email": "homer.simpson@simpsons.com",
                    "gid": "1234567890",
                    "name": "Homer Simpson",
                },
                "context": {
                    "client_ip_address": "1.2.3.4",
                    "context_type": "web",
                    "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
                },
                "created_at": "2022-12-16 19:32:00.922",
                "details": {},
                "event_category": "admin_settings",
                "event_type": "workspace_form_link_authentication_required_disabled",
                "gid": "1234567890",
                "resource": {"gid": "111234", "name": "Simpsons Lab", "resource_type": "workspace"},
            },
        ),
    ]
